# $Id: adu.bro,v 1.2 2005/10/20 02:25:35 vern Exp $

@load conn-id

module adu;

# This script parses application-layer data (ADU) units, or "messages",
# out of the packet streams.  Since the analysis is generic, we define
# an ADU simply as all application-layer data in a 5-tuple flow going
# in one direction without any data going the other way.  Once we see
# data in the other direction, we finish the current ADU and start
# a new one (going the other way).  While this approach is only
# approximate, it can work well for both UDP and TCP.
#
# The script reports ADUs as strings, up to a configurable maximum size, and
# up to a configurable depth into the flow.
#
# Generated events:
#
# - adu_tx(c: connection, a: adu_state) reports an ADU seen from
#   c's originator to its responder.
#
# - adu_rx(c: connection, a: adu_state) reports an ADU seen from
#   c's responder to the originator.


# --- Input configuration -- which ports to look at --------------------

# XXX  This isn't really what we want -- we'd like a port *range* that'd
# XXX  say "20/tcp - 8008/tcp except tcp/22", for example.  But this suffices
# XXX  for now.
redef tcp_content_delivery_ports_orig += {
	[20/tcp] = T, [21/tcp] = T, [23/tcp] = T, [25/tcp] = T,
	[43/tcp] = T, [53/tcp] = T, [79/tcp] = T, [80/tcp] = T,
	[110/tcp] = T, [111/tcp] = T, [137/tcp] = T, [138/tcp] = T,
	[139/tcp] = T, [143/tcp] = T, [513/tcp] = T, [6000/tcp] = T,
	[6001/tcp] = T, [8080/tcp] = T, [8008/tcp] = T,
};

redef tcp_content_delivery_ports_resp += { tcp_content_delivery_ports_orig };


# --- Debugging -- should really be a separate policy ------------------

# Comment out to disable debugging output:
global adu_debug = T;

# Uncomment to enable tests:
global adu_test = T;

@ifdef (adu_debug)
function DBG(msg: string) { print fmt("DBG[adu.bro]: %s", msg); }
@else
function DBG(msg: string) { }
@endif

@ifdef (adu_test)
global adu_log: file = open("adu_test.log");
@endif

# --- Constants --------------------------------------------------------

# The maximum depth in bytes up to which we follow a flow.
const ADU_CONN_MAX_DEPTH = 100000 &redef;

# The maximum message size in bytes that we report.
const ADU_MAX_SIZE = 10000 &redef;


# --- Types ------------------------------------------------------------

export {
	# adu_state records contain the latest ADU and aditional flags to help
	# the user identify the direction of the message, its depth in the flow,
	# etc.
	type adu_state: record {
		adu: string &default = "";	# the current ADU

		# Counter (>= 1) for messages in orig->resp direction.
		depth_tx: count &default = 1;

		# Counter (>= 1) for messages in resp->orig direction.
		depth_rx: count &default = 1;

		# Total size of the connection, in bytes.
		size: count &default = 0;

		# Whether this ADU is sent by the originator.
		is_orig: bool &default = F;
	};
}


# --- Globals ----------------------------------------------------------

# A global table that tracks each flow's messages.
global adu_conns: table[conn_id] of adu_state;


# --- Functions --------------------------------------------------------

function adu_copy(astate: adu_state): adu_state
	{
	return *astate;
	}


# --- Event Handlers ---------------------------------------------------

event tcp_contents(c: connection, is_orig: bool, seq: count, contents: string)
	{
	local astate: adu_state;

	DBG(fmt("contents %s, %s: %s", id_string(c$id), is_orig, contents));

	# Ensure we track the given connection.
	if ( c$id !in adu_conns )
		adu_conns[c$id] = astate;
	else
		astate = adu_conns[c$id];

	# Don't report if flow is too big.
	if ( astate$size >= ADU_CONN_MAX_DEPTH )
		return;

	# If we have an assembled message, we may now have something
	# to report.
	if ( |astate$adu| > 0 )
		{
		# If application-layer data flow is switching
		# from resp->orig to orig->resp, report the assembled
		# message as a received ADU.
		if ( is_orig && ! astate$is_orig )
			{
			event adu_rx(c, adu_copy(astate));
			astate$adu = "";
			++astate$depth_rx;
			}

		# If application-layer data flow is switching
		# from orig->resp to resp->orig, report the assembled
		# message as a transmitted ADU.
		if ( ! is_orig && astate$is_orig )
			{
			event adu_tx(c, adu_copy(astate));
			astate$adu = "";
			++astate$depth_tx;
			}
		}

	# Append the contents to the end of the currently
	# assembled message, if the message hasn't already
	# reached the maximum size.
	if ( |astate$adu| < ADU_MAX_SIZE ) {
		astate$adu += contents;
	}

	astate$size += |contents|;
	astate$is_orig = is_orig;
	}

event connection_state_remove(c: connection)
	{
	if ( c$id !in adu_conns )
		return;

	local astate = adu_conns[c$id];

	# Report the remaining data now, if any.
	if ( |astate$adu| == 0 )
		return;

	if ( astate$is_orig )
		event adu_tx(c, adu_copy(astate));
	else
		event adu_rx(c, adu_copy(astate));

	delete adu_conns[c$id];
	}


# --- Tests ------------------------------------------------------------

@ifdef (adu_test)

event adu_tx(c: connection, astate: adu_state)
	{
	print adu_log, fmt("---- %s, %d -> ----",
				id_string(c$id), astate$depth_tx);
	print adu_log, astate$adu;
	}

event adu_rx(c: connection, astate: adu_state)
	{
	print adu_log, fmt("---- %s, %d <- ----",
				id_string(c$id), astate$depth_rx);
	print adu_log, astate$adu;
	}

@endif
