// $Id: DCE_RPC.cc,v 1.3 2005/09/09 22:05:34 vern Exp $
//
// Copyright (c) 1996, 1997, 1998, 1999, 2000, 2001, 2002
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "config.h"

#include <stdlib.h>
#include <string>
#include <map>

using namespace std;

#include "DCE_RPC.h"
#include "Sessions.h"

#define xbyte(b, n) (((const u_char*) (b))[n])

#define extract_uint16(little_endian, bytes) \
	((little_endian) ? \
	 uint16(xbyte(bytes, 0)) | ((uint16(xbyte(bytes, 1))) << 8) : \
	 uint16(xbyte(bytes, 1)) | ((uint16(xbyte(bytes, 0))) << 8))

static int uuid_index[] = {
	3, 2, 1, 0,
	5, 4, 7, 6,
	8, 9, 10, 11,
	12, 13, 14, 15
};

const char* uuid_to_string(const u_char* uuid_data)
	{
	static char s[1024];
	char* sp = s;

	for ( int i = 0; i < 16; ++i )
		{
		if ( i == 4 || i == 6 || i == 8 || i == 10 )
			sp += snprintf(sp, s + sizeof(s) - sp, "-");

		int j = uuid_index[i];
		sp += snprintf(sp, s + sizeof(s) - sp, "%02x", uuid_data[j]);
		}

	return s;
	}

UUID::UUID()
	{
	memset(data, 0, 16);
	s = uuid_to_string(data);
	}

UUID::UUID(const u_char d[16])
	{
	memcpy(data, d, 16);
	s = uuid_to_string(data);
	}

UUID::UUID(const BinPAC::uuid *uuid)
	{
	memcpy(data, uuid->val().begin(), 16);
	s = uuid_to_string(data);
	}

UUID::UUID(const char* str)
	{
	s = string(str);
	const char* sp = str;
	int i;
	for ( i = 0; i < 16; ++i )
		{
		if ( *sp == '-' )
			++sp;
		if ( ! *sp || ! *(sp+1) )
			break;

		data[uuid_index[i]] =
			(u_char) (decode_hex(*sp) * 16 + decode_hex(*(sp+1)));
		}

	if ( i != 16 )
		internal_error(fmt("invalid UUID string: %s", str));
	}

typedef map<UUID, BroEnum::dce_rpc_if_id> uuid_map_t;

static uuid_map_t& well_known_uuid_map()
	{
	static uuid_map_t the_map;
	static bool initialized = false;

	if ( initialized )
		return the_map;

	using namespace BroEnum;

	the_map[UUID("e1af8308-5d1f-11c9-91a4-08002b14a0fa")] = DCE_RPC_epmapper;

	// It's said that the following interfaces are merely aliases.
	the_map[UUID("12345778-1234-abcd-ef00-0123456789ab")] = DCE_RPC_lsarpc;
	the_map[UUID("12345678-1234-abcd-ef00-01234567cffb")] = DCE_RPC_netlogon;
	the_map[UUID("12345778-1234-abcd-ef00-0123456789ac")] = DCE_RPC_samr;

	// The next group of aliases.
	the_map[UUID("4b324fc8-1670-01d3-1278-5a47bf6ee188")] = DCE_RPC_srvsvc;
	the_map[UUID("12345678-1234-abcd-ef00-0123456789ab")] = DCE_RPC_spoolss;
	the_map[UUID("45f52c28-7f9f-101a-b52b-08002b2efabe")] = DCE_RPC_winspipe;
	the_map[UUID("6bffd098-a112-3610-9833-46c3f87e345a")] = DCE_RPC_wkssvc;

	// DRS - NT directory replication service.
	the_map[UUID("e3514235-4b06-11d1-ab04-00c04fc2dcd2")] = DCE_RPC_drs;

	// "The IOXIDResolver RPC interface (formerly known as
	// IObjectExporter) is remotely used to reach the local object
	// resolver (OR)."
	the_map[UUID("99fcfec4-5260-101b-bbcb-00aa0021347a")] = DCE_RPC_oxid;

	initialized = true;
	return the_map;
	}

// Used to remember mapped DCE/RPC endpoints and parse the follow-up
// connections as DCE/RPC sessions.
map<dce_rpc_endpoint_addr, UUID> dce_rpc_endpoints;

static bool is_mapped_dce_rpc_endpoint(const dce_rpc_endpoint_addr &addr)
	{
	return dce_rpc_endpoints.find(addr) != dce_rpc_endpoints.end();
	}

bool is_mapped_dce_rpc_endpoint(const ConnID *id, TransportProto proto)
	{
	dce_rpc_endpoint_addr addr;
	addr.addr = ntohl(to_v4_addr(id->dst_addr));
	addr.port = ntohs(id->dst_port);
	addr.proto = proto;
	return is_mapped_dce_rpc_endpoint(addr);
	}

static void add_dce_rpc_endpoint(
		const dce_rpc_endpoint_addr &addr,
		const UUID &uuid)
	{
	DEBUG_MSG("Adding endpoint %s @ %s\n",
		uuid.to_string(), addr.to_string().c_str());
	dce_rpc_endpoints[addr] = uuid;
	}

DCE_RPC_Header::DCE_RPC_Header(Connection* c, const u_char* b)
	{
	conn = c;
	bytes = b;

	// This checks whether it's both the first fragment *and*
	// the last fragment.
	if ( bytes[3] & 0x3 != 0x3 )
		{
		fragmented = 1;
		Weird("Fragmented DCE/RPC message");
		}
	else
		fragmented = 0;

	ptype = (BroEnum::dce_rpc_ptype) bytes[2];
	frag_len = extract_uint16(LittleEndian(), bytes + 8);
	}

DCE_RPC_Session::DCE_RPC_Session(Connection* c)
: conn(c),
  if_uuid("00000000-0000-0000-0000-000000000000"),
  if_id(BroEnum::DCE_RPC_unknown_if)
	{
	conn = c;
	opnum = -1;
	}

bool DCE_RPC_Session::LooksLikeRPC(int len, const u_char* msg)
	{
	// if ( ! is_IPC )
	//	return false;

	try
		{
		BinPAC::DCE_RPC_Header h;
		h.parse(msg, msg + len, 0);
		if ( h.rpc_vers() == 5 && h.rpc_vers_minor() == 0 )
			{
			if ( h.frag_length() == len )
				return true;
			else
				{
				DEBUG_MSG("length mismatch: %d != %d\n",
					h.frag_length(), len);
				return false;
				}
			}
		}
	catch ( const BinPAC::Exception& )
		{
		// do nothing
		}

	return false;
	}

void DCE_RPC_Session::DeliverPDU(int is_orig, int len, const u_char* data)
	{
	if ( dce_rpc_message )
		{
		val_list* vl = new val_list;
		vl->append(conn->BuildConnVal());
		vl->append(new Val(is_orig, TYPE_BOOL));
		vl->append(new EnumVal(data[2], enum_dce_rpc_ptype));
		vl->append(new StringVal(len, (const char*) data));

		conn->ConnectionEvent(dce_rpc_message, vl);
		}

	try
		{
		BinPAC::DCE_RPC_PDU pdu;
		pdu.parse(data, data + len, 0);

		switch ( pdu.header()->PTYPE() ) {
		case BinPAC::DCE_RPC_BIND:
			DeliverBind(&pdu);
			break;

		case BinPAC::DCE_RPC_REQUEST:
			DeliverRequest(&pdu);
			break;

		case BinPAC::DCE_RPC_RESPONSE:
			DeliverResponse(&pdu);
			break;
		}
		}
	catch ( const BinPAC::Exception& e )
		{
		conn->Weird(e.msg().c_str());
		}
	}

void DCE_RPC_Session::DeliverBind(const BinPAC::DCE_RPC_PDU* pdu)
	{
	BinPAC::DCE_RPC_Bind* bind = pdu->body()->bind();

	for ( int i = 0; i < bind->p_context_elem()->n_context_elem(); ++i )
		{
		BinPAC::p_cont_elem_t* elem =
			bind->p_context_elem()->p_cont_elem()[i];

		if_uuid = UUID(elem->abstract_syntax()->if_uuid()->val().begin());
		uuid_map_t::const_iterator uuid_it =
			well_known_uuid_map().find(if_uuid);

		if ( uuid_it == well_known_uuid_map().end() )
			{
#ifdef DEBUG
			conn->Weird(fmt("Unknown DCE_RPC interface %s",
				if_uuid.to_string()));
#endif
			if_id = BroEnum::DCE_RPC_unknown_if;
			}
		else
			if_id = uuid_it->second;

		if ( dce_rpc_bind )
			{
			val_list* vl = new val_list;
			vl->append(conn->BuildConnVal());
			vl->append(new StringVal(if_uuid.to_string()));
			// vl->append(new EnumVal(if_id, enum_dce_rpc_if_id));

			conn->ConnectionEvent(dce_rpc_bind, vl);
			}
		}
	}

void DCE_RPC_Session::DeliverRequest(const BinPAC::DCE_RPC_PDU* pdu)
	{
	BinPAC::DCE_RPC_Request* req = pdu->body()->request();

	opnum = req->opnum();

	if ( dce_rpc_request )
		{
		val_list* vl = new val_list;
		vl->append(conn->BuildConnVal());
		vl->append(new Val(opnum, TYPE_COUNT));
		vl->append(new StringVal(req->stub().size(),
			(const char*) req->stub().begin()));

		conn->ConnectionEvent(dce_rpc_request, vl);
		}

	switch ( if_id ) {
	case BroEnum::DCE_RPC_epmapper:
		DeliverEpmapperRequest(pdu, req);
		break;

	default:
		break;
	}
	}

void DCE_RPC_Session::DeliverResponse(const BinPAC::DCE_RPC_PDU* pdu)
	{
	BinPAC::DCE_RPC_Response* resp = pdu->body()->response();

	if ( dce_rpc_response )
		{
		val_list* vl = new val_list;
		vl->append(conn->BuildConnVal());
		vl->append(new Val(opnum, TYPE_COUNT));
		vl->append(new StringVal(resp->stub().size(),
			(const char*) resp->stub().begin()));
		conn->ConnectionEvent(dce_rpc_response, vl);
		}

	switch ( if_id ) {
	case BroEnum::DCE_RPC_epmapper:
		DeliverEpmapperResponse(pdu, resp);
		break;

	default:
		break;
	}
	}

void DCE_RPC_Session::DeliverEpmapperRequest(const BinPAC::DCE_RPC_PDU* pdu,
					const BinPAC::DCE_RPC_Request* req)
	{
	// DEBUG_MSG("Epmapper request opnum = %d\n", req->opnum());
	switch ( opnum ) {
	}
	}

void DCE_RPC_Session::DeliverEpmapperResponse(const BinPAC::DCE_RPC_PDU* pdu,
					const BinPAC::DCE_RPC_Response* resp)
	{
	// DEBUG_MSG("Epmapper request opnum = %d\n", req->opnum());
	switch ( opnum ) {
	case 3:	// Map
		DeliverEpmapperMapResponse(pdu, resp);
		break;
	}
	}


void DCE_RPC_Session::DeliverEpmapperMapResponse(const BinPAC::DCE_RPC_PDU* pdu,
					const BinPAC::DCE_RPC_Response* resp)
	{
	try
		{
		BinPAC::epmapper_map_resp epm_resp;

		epm_resp.parse(resp->stub().begin(), resp->stub().end(),
				pdu->byteorder());

		for ( unsigned int twr_i = 0;
		      twr_i < epm_resp.towers()->actual_count(); ++twr_i )
			{
			BinPAC::epm_tower* twr =
				epm_resp.towers()->towers()[twr_i]->tower();

			mapped.addr = dce_rpc_endpoint_addr();
			mapped.uuid = UUID();

			for ( int floor_i = 0; floor_i < twr->num_floors();
			      ++floor_i )
				{
				BinPAC::epm_floor* floor = twr->floors()[floor_i];

				switch ( floor->protocol() ) {
				case BinPAC::EPM_PROTOCOL_UUID:
					if ( floor_i == 0 )
						mapped.uuid = UUID(floor->lhs()->data()->uuid()->if_uuid());
					break;

				case BinPAC::EPM_PROTOCOL_TCP:
					mapped.addr.port =
						floor->rhs()->data()->tcp()->val();
					mapped.addr.proto = TRANSPORT_TCP;
					break;

				case BinPAC::EPM_PROTOCOL_UDP:
					mapped.addr.port =
						floor->rhs()->data()->udp()->val();
					mapped.addr.proto = TRANSPORT_UDP;
					break;

				case BinPAC::EPM_PROTOCOL_IP:
					mapped.addr.addr =
						floor->rhs()->data()->ip()->val();
					break;
				}
				}

			if ( mapped.addr.is_valid_addr() )
				add_dce_rpc_endpoint(mapped.addr, mapped.uuid);

			if ( epm_map_response )
				{
				val_list* vl = new val_list;
				vl->append(conn->BuildConnVal());
				vl->append(new StringVal(mapped.uuid.to_string()));
				vl->append(new PortVal(mapped.addr.port, mapped.addr.proto));
				vl->append(new AddrVal(htonl(mapped.addr.addr)));

				conn->ConnectionEvent(epm_map_response, vl);
				}
			}
		}
	catch ( const BinPAC::Exception& e )
		{
		conn->Weird(e.msg().c_str());
		}
	}

TCP_Contents_DCE_RPC::TCP_Contents_DCE_RPC(TCP_Endpoint* arg_endp,
		DCE_RPC_Session* arg_session, bool speculative)
: TCP_Contents(arg_endp, STOP_ON_GAP, PUNT_ON_PARTIAL)
	{
	session = arg_session;
	msg_buf = 0;
	buf_len = 0;
	speculation = speculative ? 0 : 1;

	Init();
	}

void TCP_Contents_DCE_RPC::Init()
	{
	// Allocate space for header.
	if ( ! msg_buf )
		{
		buf_len = DCE_RPC_HEADER_LENGTH;
		msg_buf = new u_char[buf_len];
		}

	buf_n = 0;
	msg_len = 0;
	hdr = 0;
	}

TCP_Contents_DCE_RPC::~TCP_Contents_DCE_RPC()
	{
	delete [] msg_buf;
	delete hdr;
	}

void TCP_Contents_DCE_RPC::Deliver(int seq, int len, const u_char* data)
	{
	if ( speculation == 0 ) // undecided
		{
		if ( ! DCE_RPC_Session::LooksLikeRPC(len, data) )
			speculation = -1;
		else
			speculation = 1;
		}

	if ( speculation < 0 )
		return;

	ASSERT(buf_len >= DCE_RPC_HEADER_LENGTH);
	while ( len > 0 )
		{
		if ( buf_n < DCE_RPC_HEADER_LENGTH )
			{
			while ( buf_n < DCE_RPC_HEADER_LENGTH && len > 0 )
				{
				msg_buf[buf_n] = *data;
				++buf_n; ++data; --len;
				}

			if ( buf_n < DCE_RPC_HEADER_LENGTH )
				break;
			else
				{
				if ( ! ParseHeader() )
					return;
				}
			}

		while ( buf_n < msg_len && len > 0 )
			{
			msg_buf[buf_n] = *data;
			++buf_n; ++data; --len;
			}

		if ( buf_n < msg_len )
			break;
		else
			{
			if ( msg_len > 0 )
				DeliverPDU(msg_len, msg_buf);
			// Reset for next message
			Init();
			}
		}
	}

void TCP_Contents_DCE_RPC::DeliverPDU(int len, const u_char* data)
	{
	session->DeliverPDU(IsOrig(), len, data);
	}

bool TCP_Contents_DCE_RPC::ParseHeader()
	{
	delete hdr;
	hdr = 0;

	if ( msg_buf[0] != 5 )	// DCE/RPC version
		{
		Conn()->Weird("DCE/RPC_version_error (non-DCE/RPC?)");
		Conn()->SetSkip(1);
		msg_len = 0;
		return false;
		}

	hdr = new DCE_RPC_Header(endp->Conn(), msg_buf);

	msg_len = hdr->FragLen();
	if ( msg_len > buf_len )
		{
		u_char* new_msg_buf = new u_char[msg_len];
		memcpy(new_msg_buf, msg_buf, buf_n);
		delete [] msg_buf;
		buf_len = msg_len;
		msg_buf = new_msg_buf;
		hdr->SetBytes(new_msg_buf);
		}

	return true;
	}

DCE_RPC_Conn::DCE_RPC_Conn(NetSessions* s, HashKey* k, double t,
		const ConnID* id, const struct tcphdr* tp,
		bool arg_speculative)
: TCP_Connection(s, k, t, id, tp)
	{
	o_dce_rpc = r_dce_rpc = 0;
	session = new DCE_RPC_Session(this);
	speculative = arg_speculative;
	}

DCE_RPC_Conn::~DCE_RPC_Conn()
	{
	delete session;
	}

void DCE_RPC_Conn::BuildEndpoints()
	{
	o_dce_rpc = new TCP_Contents_DCE_RPC(orig, session, speculative);
	r_dce_rpc = new TCP_Contents_DCE_RPC(resp, session, speculative);

	orig->AddContentsProcessor(o_dce_rpc);
	resp->AddContentsProcessor(r_dce_rpc);
	}
