// $Id: pac.cc,v 1.2 2005/09/07 17:33:56 vern Exp $

#include <string>

using namespace std;

#include "pac.h"
#include "pac_dbg.h"
#include "pac_exception.h"

const ID default_value_var("val");
const ID begin_of_data("begin_of_data");
const ID end_of_data("end_of_data");
const ID len_of_data("len_of_data");
const ID byteorder_id("byteorder");
const ID bigendian_id("bigendian");
const ID littleendian_id("littleendian");

Env* global_env;

typedef map<const ID *, Type*, ID_cmp> type_map_t;
type_map_t type_map;

int pac_init()
	{
	global_env = new Env(0);

	// These two are defined in binpac.h, so we do not need to
	// generate code for them.
	global_env->AddConstID(&bigendian_id, 0);
	global_env->AddConstID(&littleendian_id, 1);
	global_env->SetEvaluated(&bigendian_id);
	global_env->SetEvaluated(&littleendian_id);

	return 0;
	}

IDRecord::IDRecord(const ID* arg_id, ID_type arg_id_type)
	: id(arg_id), id_type(arg_id_type)
	{
	eval = 0;
	evaluated = in_evaluation = false;
	switch (id_type)
		{
		case MEMBER_VAR:
			rvalue = nfmt("%s()", id->Name());
			lvalue = nfmt("%s_", id->Name());
			break;
		case UNION_VAR:
			rvalue = nfmt("%s()", id->Name());
			lvalue = nfmt("%s_", id->Name());
			break;
		case CONST:
		case GLOBAL_VAR:
			rvalue = nfmt("%s", id->Name());
			lvalue = nfmt("%s", id->Name());
			break;
		case TEMP_VAR:
			rvalue = nfmt("t_%s", id->Name());
			lvalue = nfmt("t_%s", id->Name());
			break;
		}
	field = 0;
	}

IDRecord::~IDRecord()
	{
	delete [] rvalue;
	delete [] lvalue;
	}

void IDRecord::Evaluate(Output* out, Env* env)	
	{ 
	if ( evaluated )
		return;
		
	if ( in_evaluation )
		throw ExceptionCyclicDependence(id);

	in_evaluation = true;
	eval->genEval(out, env); 
	in_evaluation = false;

	evaluated = true;
	}

const char* IDRecord::RValue() const
	{
	if ( id_type == TEMP_VAR && ! evaluated )
		throw Exception(id, 
			fmt("Id '%s' not evaluated before used", id->Name()));
	return rvalue;
	}

Env::Env(Env* parent_env)
	: parent(parent_env)
	{
	}

Env::~Env()
	{
	for ( id_map_t::iterator it = id_map.begin(); it != id_map.end(); ++it )
		{
		delete it->second;
		it->second = 0;
		}
	}

void Env::AddID(const ID* id, ID_type id_type)
	{
	DEBUG_MSG("To add ID `%s'...\n", id->Name());
	id_map_t::iterator it = id_map.find(id);
	if ( it != id_map.end() )
		{
		DEBUG_MSG("Duplicate definition: `%s'\n", it->first->Name());
		throw ExceptionIDRedefinition(id);
		}
	id_map[id] = new IDRecord(id, id_type);
	}

void Env::AddConstID(const ID* id, const int c)
	{
	AddID(id, CONST);
	SetConstant(id, c);
	}

IDRecord* Env::lookup(const ID* id, bool recursive, bool raise_exception) const
	{
	id_map_t::const_iterator it = id_map.find(id);
	if ( it != id_map.end() )
		return it->second;

	if ( recursive && parent )
		return parent->lookup(id, recursive, raise_exception);

	if ( raise_exception )
		throw ExceptionIDNotFound(id);
	else
		return 0;
	}

const char* Env::RValue(const ID* id) const
	{
	return lookup(id)->RValue();
	}

const char* Env::LValue(const ID* id) const
	{
	return lookup(id)->LValue();
	}

void Env::SetEvalMethod(const ID* id, Evaluatable* eval)
	{
	lookup(id)->SetEvalMethod(eval);
	}

void Env::Evaluate(Output* out, const ID* id)
	{
	lookup(id)->Evaluate(out, this);
	}

bool Env::Evaluated(const ID* id) const
	{
	return lookup(id)->Evaluated();
	}

void Env::SetEvaluated(const ID* id, bool v)
	{
	lookup(id)->SetEvaluated(v);
	}

void Env::SetField(const ID* id, Field* field)
	{
	lookup(id, false)->SetField(field);
	}

Field* Env::getField(const ID* id) const
	{
	return lookup(id)->getField();
	}

void Env::SetConstant(const ID* id, int constant)
	{
	lookup(id, false)->SetConstant(constant);
	}

bool Env::getConstant(const ID* id, int& c) const
	{
	// lookup without raising exception
	IDRecord* r = lookup(id, true, false);
	if ( r )
		return r->getConstant(c);
	else
		return false;
	}

DataPtr::DataPtr(Env* env, const ID* arg_id, const int arg_off)
	: id(arg_id), off(arg_off)
	{
	if ( ! env->Evaluated(id) )
		throw Exception(id, 
			fmt("Id '%s' not evaluated before used", id->Name()));
	if ( off == 0 )
		ptr_expr = nfmt("%s", env->RValue(id));
	else
		ptr_expr = nfmt("(%s + %d)", env->RValue(id), off);
	}

int DataPtr::AbsOffset(const ID* base_ptr) const
	{
	return ( id == base_ptr ) ? off : -1;
	}

char* DataPtr::AbsOffsetExpr(Env* env, const ID* base_ptr) const
	{
	if ( AbsOffset(base_ptr) >= 0 )
		return nfmt("%d", off);
	else
		return nfmt("(%s - %s)", PtrExpr(), env->RValue(base_ptr));
	}

void DataPtr::genBoundaryCheck(Output* out_cc, Env* env,
		const char* data_size, const char* data_name) const
	{
	out_cc->println("// Checking out-of-bound for \"%s\"", data_name);
	out_cc->println("if ( %s + (%s) > %s )",
		PtrExpr(),
		data_size,
		env->RValue(&end_of_data));

	out_cc->inc_indent(); 
	out_cc->println("{");

	char* data_offset = AbsOffsetExpr(env, &begin_of_data); 

	out_cc->println("// Handle out-of-bound condition");
	out_cc->println("throw ExceptionOutOfBound(\"%s\",", data_name);
	out_cc->println("	(%s) + (%s), ", 
		data_offset, data_size);
	out_cc->println("	(%s) - (%s));", 
		env->RValue(&end_of_data), env->RValue(&begin_of_data));

	delete [] data_offset;

	out_cc->println("}");
	out_cc->dec_indent(); 
	}

Decl::Decl(ID* arg_id)
	: id(arg_id), attrlist(0)
	{
	}

Decl::~Decl()
	{
	delete id;
	delete_list(AttrList, attrlist);
	}

void Decl::AddAttr(AttrList* attrs)
	{
	if ( attrlist)
		throw Exception(this, "multiple AddAttr");
	attrlist = attrs;
	foreach ( i, AttrList, attrlist )
		ProcessAttr(*i);
	}

Enum::Enum(const ID* arg_id, Expr* arg_expr)
	: id(arg_id), expr(arg_expr)
	{
	}

void Enum::genHeader(Output* out_h, int& c)
	{
	if ( expr )
		{
		if ( ! expr->ConstFold(global_env, c) )
			throw ExceptionNonConstExpr(expr);
		out_h->println("%s = %d,", id->Name(), c);
		}
	else
		out_h->println("%s,", id->Name());
	global_env->AddConstID(id, c);
	global_env->SetEvaluated(id);
	}

void EnumDecl::genCode(Output* out_h, Output* /* out_cc */)
	{
	out_h->println("enum %s {", id->Name());
	out_h->inc_indent();
	int c = 0;
	foreach(i, EnumList, enumlist)
		{
		(*i)->genHeader(out_h, c);
		++c;
		}
	out_h->dec_indent();
	out_h->println("};");
	}

LetDecl::LetDecl(ID* arg_id, Expr* arg_expr)
 	: Decl(arg_id), expr(arg_expr)
	{
	int c;
	if ( expr->ConstFold(global_env, c) )
		global_env->AddConstID(id, c);
	else
		global_env->AddID(id, GLOBAL_VAR);
	}

LetDecl::~LetDecl()
	{
	delete id;
	delete expr;
	}

void LetDecl::genCode(Output* out_h, Output* /* out_cc */)
	{
	out_h->println("static const int %s = %s;", 
		global_env->LValue(id), expr->EvalExpr(out_h, global_env));
	global_env->SetEvaluated(id);
	}

Type* look_up_type(const ID* id)
	{
	type_map_t::iterator it = type_map.find(id);
	if ( it == type_map.end() )
		return 0;
	return it->second;
	}

TypeDecl::TypeDecl(ID* arg_id, ParamList* arg_params, Type* arg_type)
	: Decl(arg_id), params(arg_params), type(arg_type)
	{
	env = new Env(global_env);
	type->SetDeclID(id);
	type_map[id] = type;
	byteorder_expr = 0;
	}

TypeDecl::~TypeDecl()
	{
	delete env;
	delete type;

	delete_list(ParamList, params);

	delete id;
	}

void TypeDecl::ProcessAttr(Attr* a)
	{
	if ( a->Type() == ATTR_BYTEORDER )
		{
		AttrByteOrder* ab = (AttrByteOrder*) a;
		byteorder_expr = ab->IsLittleEndian();
		}
	else if ( a->Type() == ATTR_CHECK )
		{
		AttrCheck* ac = (AttrCheck*) a;
		type->AddCheck(ac->getExpr());
		}
	}

void TypeDecl::genCode(Output* out_h, Output* out_cc)
	{
	Prepare();

	fprintf(stderr, "Generating code for %s\n", ClassName());

	// gen the header file
	out_h->println("");
	out_h->println("class %s", ClassName());
	out_h->println("{");
	out_h->println("public:");
	out_h->inc_indent();

	genConstructorFunc(out_h, out_cc);
	genDestructorFunc(out_h, out_cc);
	genParseFunc(out_h, out_cc);

	out_h->println("");
	out_h->println("// Member access functions");

	genParamPubDecls(out_h);
	type->genPubDecls(out_h, env);
	out_h->println("");

	out_h->dec_indent();
	out_h->println("protected:");
	out_h->inc_indent();

	genParamPrivDecls(out_h);
	type->genPrivDecls(out_h, env);

	out_h->dec_indent();
	out_h->println("};\n");
	}

void TypeDecl::Prepare()
	{
	foreach (i, ParamList, params)
		{
		Param* p = *i;
		env->AddID(p->getID(), MEMBER_VAR);
		}

	if ( byteorder_expr )
		env->AddID(&byteorder_id, MEMBER_VAR);

	type->Prepare(env, 0);
	}

void TypeDecl::genParamPubDecls(Output* out_h)
	{
	foreach (i, ParamList, params)
		{
		Param* p = *i;
		out_h->println("%s const %s const { return %s; };", 
			p->getTypeName(), 
			env->RValue(p->getID()), 
			env->LValue(p->getID())); 
		}

	if ( byteorder_expr )
		out_h->println("int const %s const { return %s; };",
			env->RValue(&byteorder_id), 
			env->LValue(&byteorder_id)); 
	}

void TypeDecl::genParamPrivDecls(Output* out_h)
	{
	foreach (i, ParamList, params)
		{
		Param* p = *i;
		out_h->println("%s %s;", 
			p->getTypeName(), 
			env->LValue(p->getID()));
		}

	if ( byteorder_expr )
		out_h->println("int %s;",
			env->LValue(&byteorder_id)); 
	}

char* TypeDecl::ConstructorFuncPrototype(Env* env)
	{
	string param_decls;
	const char* param_str = "";

	int first = 1;
	foreach (i, ParamList, params)
			{
			Param* p = *i;
			const char* decl_str = p->Decl();
			if ( first )
				first = 0;
			else
				param_decls += ", ";
			param_decls += decl_str;
			}
		param_str = param_decls.c_str();
	return nfmt("%s(%s)", id->Name(), param_str);
	}

void TypeDecl::genConstructorFunc(Output* out_h, Output* out_cc)
	{
	char* proto = ConstructorFuncPrototype(env);

	out_h->println("%s;", proto);

	out_cc->println("%s::%s", ClassName(), proto);
	out_cc->inc_indent();
	out_cc->println("{");

	foreach (i, ParamList, params)
		{
		Param* p = *i;
		out_cc->println("%s = arg_%s;", 
			env->LValue(p->getID()), p->Name());
		env->SetEvaluated(p->getID());
		}

	type->genInitCode(out_cc, env);

	out_cc->println("}\n");
	out_cc->dec_indent();

	delete [] proto;
	}

char* TypeDecl::DestructorFuncPrototype(Env* env)
	{
	return nfmt("~%s()", id->Name());
	}

void TypeDecl::genDestructorFunc(Output* out_h, Output* out_cc)
	{
	char* proto = DestructorFuncPrototype(env);

	out_h->println("%s;", proto);

	out_cc->println("%s::%s", ClassName(), proto);
	out_cc->inc_indent();
	out_cc->println("{");

	type->genCleanUpCode(out_cc, env);

	out_cc->println("}\n");
	out_cc->dec_indent();

	delete [] proto;
	}

char* TypeDecl::ParseFuncPrototype(Env* env)
	{
	Env e(env);
	e.AddID(&byteorder_id, TEMP_VAR);
	return nfmt("parse(const uint8* const %s, const uint8* const %s, int %s)", 
		e.LValue(&begin_of_data), 
		e.LValue(&end_of_data), 
		e.LValue(&byteorder_id));
	}

void TypeDecl::genParseFunc(Output* out_h, Output* out_cc)
	{
	// Env within the parse function
	Env p_func_env(env);

	p_func_env.AddID(&begin_of_data, TEMP_VAR);
	p_func_env.AddID(&end_of_data, TEMP_VAR);

	// Do we have a custom byteorder specified for this type?
	// (Or inherited from the upper level parsing?)
	LetDef* custom_byteorder = 0;

	if ( byteorder_expr )
		{
		custom_byteorder = new LetDef(&byteorder_id, byteorder_expr);
		p_func_env.SetEvalMethod(&byteorder_id, custom_byteorder);
		}
	else if ( type->ByteOrderSensitive() )
		{
		p_func_env.AddID(&byteorder_id, TEMP_VAR);
		p_func_env.SetEvaluated(&byteorder_id);
		}

	char* proto = ParseFuncPrototype(&p_func_env);

	out_h->println("int %s;", proto);

	out_cc->println("int %s::%s", ClassName(), proto);
	out_cc->inc_indent();
	out_cc->println("{");

	p_func_env.SetEvaluated(&begin_of_data);
	p_func_env.SetEvaluated(&end_of_data);

	DataPtr data(&p_func_env, &begin_of_data, 0);
	type->genParseCode(out_cc, &p_func_env, data);

	out_cc->println("return %s;", type->DataSize(0, &p_func_env, data));

	out_cc->println("}\n");
	out_cc->dec_indent();

	delete custom_byteorder;
	delete [] proto;
	}

Param::Param(ID* arg_id, ID* arg_type_id)
	: id(arg_id), type_id(arg_type_id)
	{
	type_name = type_id ?
			nfmt("%s*", type_id->Name()) :
			nfmt("int");
	decl_str = nfmt("%s arg_%s", type_name, id->Name());
	}

Param::~Param()
	{
	delete type_name;
	delete decl_str;
	}

Type::Type() 
	{ 
	tot_ = UNDEF; 
	decl_id = current_decl_id; 
	data_id_str = 0;
	value_var = &default_value_var; 
	value_var_type = MEMBER_VAR;
	size_var = 0;
	size_expr = 0;
	boundary_checked = false;
	lvalue = 0;
	}

Type::~Type()
	{
	delete [] data_id_str;
	delete size_var;
	delete [] size_expr;
	delete [] lvalue;
	}

void Type::Prepare(Env* env, const char* enclosing_lvalue)
	{
	env->AddID(value_var, value_var_type);
	lvalue = enclosing_lvalue ?
		nfmt("%s.%s", enclosing_lvalue, env->LValue(value_var)) :
		nfmt("%s", env->LValue(value_var));
	data_id_str = nfmt("%s:%s", decl_id->Name(), value_var->Name());
	}

void Type::genPubDecls(Output* out_h, Env* env)
	{
	if ( DataType() )
		out_h->println("%s const & %s const	{ return %s; }", 
			DataType(), 
			env->RValue(value_var), lvalue);
	}

void Type::genPrivDecls(Output* out_h, Env* env)
	{
	if ( DataType() )
		out_h->println("%s %s;", 
			DataType(), 
			env->LValue(value_var));
	}

void Type::genInitCode(Output* out_cc, Env* env)
	{
	if ( DataType() )
		out_cc->println("%s = 0;", lvalue);
	}

void Type::genCleanUpCode(Output* out_cc, Env* env)
	{
	// nothing
	}

// A wrapper for do_genParseCode, which does the real job
void Type::genParseCode(Output* out_cc, Env* env, const DataPtr& data)
	{
	if ( env->Evaluated(value_var) )
		return;

	do_genParseCode(out_cc, env, data);

	env->SetEvaluated(value_var);
	}

// Returns false if size_var is already added
bool Type::AddSizeVar(Output* out_cc, Env* env)
	{
	if ( StaticSize(env) >= 0 )
		return false;

	if ( size_var )
		return false;

	ASSERT ( value_var != 0 );
	DEBUG_MSG("adding size var for `%s'\n", value_var->Name());
	size_var = new ID(fmt("%s__size", value_var->Name()));
	env->AddID(size_var, TEMP_VAR);
	// out_cc->println("int %s;", env->LValue(size_var));
	return true;
	}

const char* Type::DataSize(Output* out_cc, Env* env, const DataPtr& data)
	{
	if ( size_expr )
		return size_expr;

	int ss = StaticSize(env);
	if ( ss >= 0 )
		{
		size_expr = nfmt("%d", ss);
		return size_expr;
		}
	else
		{
		if ( ! size_var || ! env->Evaluated(size_var) )
			{
			ASSERT(out_cc != 0);
			genDynamicSize(out_cc, env, data);
			}
		return env->RValue(size_var);
		}
	}

void Type::genBoundaryCheck(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	if ( boundary_checked )
		return;

	data.genBoundaryCheck(out_cc, env, 
		DataSize(out_cc, env, data),
		data_id_str);

	SetBoundaryChecked();
	}

bool Type::needsCleanUp() const
	{
	switch ( tot_ )
		{
		case PARAMETERIZED:
		case ARRAY:
		case SEQ:
			return true;
		case BASIC:
			return false;
		default:
			ASSERT(0);
			return true;
		}
	return true;
	}

static const char* basic_pactype_name[] = {
#	define TYPE_DEF(name, pactype, ctype, size)	pactype,
#	include "pac_type.def"
#	undef TYPE_DEF
	0,
};

int BasicType::lookup_basictype(const char* name)
	{
	for ( int i = 0; basic_pactype_name[i]; ++i )
		if ( strcmp(basic_pactype_name[i], name) == 0 )
			return i;
	return -1;
	}

static const char* basic_ctype_name[] = {
#	define TYPE_DEF(name, pactype, ctype, size)	ctype,
#	include "pac_type.def"
#	undef TYPE_DEF
	0,
};

const char* BasicType::DataType()
	{
	return basic_ctype_name[type];
	}

int BasicType::StaticSize(Env* /* env */) const
	{
	static const size_t basic_type_size[] = 
		{
#		define TYPE_DEF(name, pactype, ctype, size)	size,
#		include "pac_type.def"
#		undef TYPE_DEF
		};

	return basic_type_size[type];
	}

void BasicType::genDynamicSize(Output* out_cc, Env* env, const DataPtr& data)
	{
	/* should never be called */
	ASSERT(0);
	}

void BasicType::do_genParseCode(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	// There is no need to generate the size variable
	// out_cc->println("%s = sizeof(%s);", size_var, DataType());

	genBoundaryCheck(out_cc, env, data);

	switch ( type ) 
		{
		case EMPTY:
			// do nothing
			break;

		case INT8:
		case UINT8:
			out_cc->println("%s = *((%s *) (%s));",
				lvalue, DataType(), data.PtrExpr()); 
			break;
		case INT16:
		case UINT16:
		case INT32:
		case UINT32:
			env->Evaluate(out_cc, &byteorder_id);
			out_cc->println("%s = UnMarshall<%s>(%s, %s, %s);",
				lvalue, 
				DataType(), 
				data.PtrExpr(),
				lvalue, 
				env->RValue(&byteorder_id)); 
			break;
		}
	}

ParameterizedType::ParameterizedType(ID* arg_type_id, ExprList* arg_args)
	: type_id(arg_type_id), args(arg_args)
	{
	tot_ = PARAMETERIZED;
	data_type = nfmt("%s *", type_id->Name());
	}

ParameterizedType::~ParameterizedType()
	{
	delete [] data_type;
	}

const char* ParameterizedType::DataType()
	{
	return data_type;
	}

int ParameterizedType::StaticSize(Env* env) const
	{
	Type* type = look_up_type(type_id);
	if ( ! type )
		throw ExceptionIDNotFound(type_id);

	return type->StaticSize(env);
	}

bool ParameterizedType::ByteOrderSensitive() const
	{
	Type* type = look_up_type(type_id);
	if ( ! type )
		throw ExceptionIDNotFound(type_id);

	return type->ByteOrderSensitive();
	}

void ParameterizedType::genCleanUpCode(Output* out_cc, Env* env)
	{
	out_cc->println("delete %s;", lvalue);
	}

void ParameterizedType::do_genParseCode(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	string arg_str;

	int first = 1;
	foreach (i, ExprList, args)
		{
		Expr* e = *i;
		if ( first )
			first = 0;
		else
			arg_str += ", ";
		arg_str += e->EvalExpr(out_cc, env);
		}

	out_cc->println("%s = new %s(%s);", 
		lvalue, type_id->Name(), arg_str.c_str());

	const char* byteorder_str;
	if ( ByteOrderSensitive() )
		{
		env->Evaluate(out_cc, &byteorder_id);
		byteorder_str = env->RValue(&byteorder_id);
		}
	else
		byteorder_str = "bigendian";

	AddSizeVar(out_cc, env);
	bool compute_size_var = (size_var && ! env->Evaluated(size_var));
	if ( compute_size_var )
		{
		out_cc->println("int %s = ", env->LValue(size_var));
		out_cc->inc_indent();
		}

	out_cc->println("%s->parse(%s, %s, %s);",
		lvalue, // parse() needs an LValue
		data.PtrExpr(), 
		env->RValue(&end_of_data),
		byteorder_str);

	if ( compute_size_var )
		{
		out_cc->dec_indent();
		env->SetEvaluated(size_var);
		}
	}

void ParameterizedType::genDynamicSize(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	genParseCode(out_cc, env, data);
	}

ArrayType::ArrayType(Type* arg_elemtype, Expr* arg_length)
	: elemtype(arg_elemtype), length(arg_length), end(0)
	{
	init();
	// data_type = nfmt("%s *", elemtype->DataType());
	}

ArrayType::ArrayType(Type* arg_elemtype, SeqEnd* arg_end)
	: elemtype(arg_elemtype), length(0), end(arg_end)
	{
	init();
	}

void ArrayType::init()
	{
	tot_ = ARRAY;
	arraylength_var = 0;
	elem_var = 0;
	elem_data_var = 0;
	data_type = nfmt("array<%s>", elemtype->DataType());
	}

ArrayType::~ArrayType()
	{
	delete elem_var;
	delete elem_data_var;
	delete data_type;
	delete arraylength_var;
	}

const char* ArrayType::DataType()
	{
	return data_type;
	}

void ArrayType::Prepare(Env* env, const char* enclosing_lvalue)
	{
	Type::Prepare(env, enclosing_lvalue);

	arraylength_var = new ID(fmt("%s__arraylength", value_var->Name()));
	env->AddID(arraylength_var, TEMP_VAR);

	elem_var = new ID(fmt("%s__elem", value_var->Name()));
	elemtype->SetValueVar(elem_var, TEMP_VAR);
	elemtype->Prepare(env, 0);

	elem_data_var = new ID(fmt("%s__data", elem_var->Name()));
	env->AddID(elem_data_var, TEMP_VAR);
	}

void ArrayType::genArrayLength(Output* out_cc, Env* env)
	{
	if ( ! length )
		return;

	if ( env->Evaluated(arraylength_var) )
		return;

	out_cc->println("int %s = %s;", 
		env->LValue(arraylength_var), length->EvalExpr(out_cc, env));

	env->SetEvaluated(arraylength_var);
	}

void ArrayType::genPubDecls(Output* out_h, Env* env)
	{
	Type::genPubDecls(out_h, env);
	}

void ArrayType::genPrivDecls(Output* out_h, Env* env)
	{
	Type::genPrivDecls(out_h, env);
	}

void ArrayType::genInitCode(Output* out_cc, Env* env)
	{
	out_cc->println("make_array(%s, 0);", lvalue);
	// Type::genInitCode(out_cc, env);
	}

void ArrayType::genCleanUpCode(Output* out_cc, Env* env)
	{
	if ( elemtype->needsCleanUp() )
		{
		out_cc->println("for ( int i = 0; i < %s.size(); ++i )", 
			env->RValue(value_var));
		out_cc->inc_indent();
		out_cc->println("{");
		out_cc->println("%s %s = %s[i];", 
			elemtype->DataType(), 
			env->LValue(elem_var), 
			lvalue);
		elemtype->genCleanUpCode(out_cc, env);
		out_cc->println("}");
		out_cc->dec_indent();
		}
	out_cc->println("delete_array(%s);", lvalue);
	}

void ArrayType::do_genParseCode(Output* out_cc, Env* env, 
		const DataPtr& data)
	{
	bool compute_size_var;

	if ( ! end && 
             ( StaticSize(env) >= 0 || elemtype->StaticSize(env) >= 0 ) )
		{
		compute_size_var = false;
		genBoundaryCheck(out_cc, env, data);
		}
	else
		{
		AddSizeVar(out_cc, env);
		compute_size_var = (size_var && ! env->Evaluated(size_var));
		}

	genArrayLength(out_cc, env);
	char* array_str;
	bool use_vector;
	if ( env->Evaluated(arraylength_var) )
		{
		out_cc->println("make_array<%s>(%s, %s);",
			elemtype->DataType(), 
			lvalue,
			env->RValue(arraylength_var));
		array_str = nfmt("%s.v", lvalue);
		use_vector = false;
		}
	else
		{
		array_str = nfmt("t_%s__vec", value_var->Name());
		out_cc->println("std::vector<%s> %s;", 
			elemtype->DataType(),
			array_str);
		use_vector = true;
		}

	char* elem_it = nfmt("t_%s__it", elem_var->Name());

	out_cc->println("const uint8* %s = %s;", 
		env->LValue(elem_data_var), data.PtrExpr());
	env->SetEvaluated(elem_data_var);

	DataPtr elem_data(env, elem_data_var, 0);

	out_cc->println("for ( int %s = 0; ", elem_it);
	if ( env->Evaluated(arraylength_var) )
		out_cc->println("	%s < %s; ", elem_it, env->RValue(arraylength_var));
	else
		out_cc->println("	/* forever */; ");
	out_cc->println("	++%s )", elem_it);
	out_cc->inc_indent();
	out_cc->println("{");
	if ( end )
		end->genParseCode(out_cc, env, elem_data_var, value_var->Name());
	out_cc->println("%s %s;", 
		elemtype->DataType(), env->LValue(elem_var));

	elemtype->genParseCode(out_cc, env, elem_data);

	// Assign the element
	if ( ! use_vector )
		out_cc->println("%s[%s] = %s;", 
			array_str, elem_it, env->RValue(elem_var));
	else
		out_cc->println("%s.push_back(%s);", 
			array_str, env->RValue(elem_var));

	out_cc->println("%s += %s;", 
		env->LValue(elem_data_var), elemtype->DataSize(0, env, elem_data));
	out_cc->println("}");
	out_cc->dec_indent();

	if ( use_vector )
		{
		out_cc->println("make_array<%s>(%s, %s.size());",
			elemtype->DataType(), 
			lvalue,
			array_str);
		out_cc->println("for ( int %s = 0; %s < %s.length; ++%s )", 
			elem_it, elem_it, env->RValue(value_var), elem_it);
		out_cc->inc_indent();
		out_cc->println("%s[%s] = %s[%s];", 
			lvalue, elem_it,
			array_str, elem_it);
		out_cc->dec_indent();
		}

	if ( compute_size_var )
		{
		// Compute the data size
		out_cc->println("int %s = %s - %s;", 
			env->LValue(size_var),
			env->RValue(elem_data_var),
			data.PtrExpr());
		env->SetEvaluated(size_var);
		}

	delete [] array_str;
	delete [] elem_it;

	elemtype->SetValueVar(0, CONST);
	}

void ArrayType::genDynamicSize(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	DEBUG_MSG("Generating dynamic size for array `%s'\n", 
		value_var->Name());

	int elem_w = elemtype->StaticSize(env);
	if ( ! end && length && elem_w >= 0 )
		{
		// If the elements have a fixed size,
		// we can get away by only dynamically computing the array length
		AddSizeVar(out_cc, env);
		bool compute_size_var = (size_var && ! env->Evaluated(size_var));
		ASSERT(compute_size_var);
		genArrayLength(out_cc, env);
		ASSERT(env->Evaluated(arraylength_var));
		out_cc->println("int %s = %d * %s;",
			env->LValue(size_var), elem_w, env->RValue(arraylength_var));
		env->SetEvaluated(size_var);
		return;
		}
	else
		// Otherwise we need parse the array dynamically
		genParseCode(out_cc, env, data);
	}

int ArrayType::StaticSize(Env* env) const
	{
	int num = 0;

	if ( end )
		return -1;

	if ( ! length || ! length->ConstFold(env, num) )
		return -1;

	int elem_w = elemtype->StaticSize(env);
	if ( elem_w < 0 )
		return -1;

	DEBUG_MSG("static size of %s:%s = %d * %d\n", 
		decl_id->Name(), lvalue, elem_w, num);

	return num * elem_w;
	}

void ArrayType::SetBoundaryChecked()
	{
	Type::SetBoundaryChecked();
	elemtype->SetBoundaryChecked();
	}

RecordType::RecordType(FieldList* arg_fields)
	: fields(arg_fields) 
	{
	tot_ = RECORD;
	}

RecordType::~RecordType()
	{
	}

const char* RecordType::DataType()
	{
	ASSERT(0);
	return 0;
	}

void RecordType::Prepare(Env* env, const char* enclosing_lvalue)
	{
	Type::Prepare(env, enclosing_lvalue);

	Field* prev = 0;
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		f->SetRecordType(this);
		f->Prepare(env);
		f->SetPrev(prev);
		if ( prev )
			prev->SetNext(f);
		prev = f;
		}
	}

void RecordType::genPubDecls(Output* out_h, Env* env)
	{
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		f->genPubDecls(out_h, env);
		}
	}

void RecordType::genPrivDecls(Output* out_h, Env* env)
	{
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		f->genPrivDecls(out_h, env);
		}
	}

void RecordType::genInitCode(Output* out_cc, Env* env)
	{
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		f->genInitCode(out_cc, env);
		}
	}

void RecordType::genCleanUpCode(Output* out_cc, Env* env)
	{
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		f->genCleanUpCode(out_cc, env);
		}
	}

void RecordType::do_genParseCode(Output* out_cc, Env* env, 
		const DataPtr& data)
	{
	if ( StaticSize(env) >= 0 )
		genBoundaryCheck(out_cc, env, data);

	ASSERT(	data.getID() == &begin_of_data && 
		data.getOffset() == 0 );

	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		f->genParseCode(out_cc, env);
		out_cc->println("");
		}

	if ( AddSizeVar(out_cc, env) )
		{
		const DataPtr& end_of_record_dataptr = fields->back()->getFieldEnd(out_cc, env);
		out_cc->println("int %s = %s - %s;", 
			env->LValue(size_var), 
			end_of_record_dataptr.PtrExpr(), env->RValue(&begin_of_data));
		env->SetEvaluated(size_var);
		}

	if ( ! boundary_checked )
		{
		Field* last_field = fields->back();
		if ( ! last_field->BoundaryChecked() )
			genBoundaryCheck(out_cc, env, data);
		}
	}

void RecordType::genDynamicSize(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	genParseCode(out_cc, env, data);
	}

int RecordType::StaticSize(Env* env) const
	{
	int tot_w = 0;
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		int w = f->StaticSize(env, tot_w);
		if ( w < 0 )
			return -1;
		tot_w += w;
		}
	return tot_w;
	}

void RecordType::SetBoundaryChecked()
	{
	Type::SetBoundaryChecked();
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		f->SetBoundaryChecked();
		}
	}

bool RecordType::ByteOrderSensitive() const
	{
	foreach (i, FieldList, fields)
		{
		Field* f = *i;
		if ( f->ByteOrderSensitive() )
			return true;
		}
	return false;
	}

Field::Field(ID* arg_id, FieldType tof)
	: id(arg_id), tof_(tof)
	{
	begin_of_field_dataptr = 0;
	end_of_field_dataptr = 0;
	field_size_expr = 0;
	field_offset_expr = 0;
	end_of_field_dataptr_var = 0;
	record_type = 0;
	prev = 0;
	next = 0;
	boundary_checked = false;
	attrlist = 0;
	decl_id = current_decl_id;
	field_id_str = nfmt("%s:%s", decl_id->Name(), id->Name());
	}

Field::~Field()
	{
	delete begin_of_field_dataptr;
	delete end_of_field_dataptr;
	delete [] field_size_expr;
	delete [] field_offset_expr;
	delete end_of_field_dataptr_var;
	delete_list(AttrList, attrlist);
	delete [] field_id_str;
	}

void Field::AddAttr(AttrList* attrs)
	{
	if ( attrlist)
		throw Exception(this, "multiple AddAttr");
	attrlist = attrs;
	}

const DataPtr& Field::getFieldBegin(Output* out_cc, Env* env)
	{
	if ( prev )
		return prev->getFieldEnd(out_cc, env);
	else
		{
		// The first field
		if ( ! begin_of_field_dataptr )
			begin_of_field_dataptr = new DataPtr(env, &begin_of_data, 0);
		return *begin_of_field_dataptr;
		}	
	}

const DataPtr& Field::getFieldEnd(Output* out_cc, Env* env)
	{
	if ( end_of_field_dataptr )
		return *end_of_field_dataptr;

	const DataPtr& begin_ptr = getFieldBegin(out_cc, env);

	int field_offset;
	if ( begin_ptr.getID() == &begin_of_data )
		field_offset = begin_ptr.getOffset();
	else
		field_offset = -1;	// unknown
		
	int field_size = StaticSize(env, field_offset);
	if ( field_size >= 0 ) // can be statically determinted 
		{
		end_of_field_dataptr = new DataPtr(
			env,
			begin_ptr.getID(), 
			begin_ptr.getOffset() + field_size);
		return *end_of_field_dataptr;
		}

	// If not, we add a variable for the offset after the field
	end_of_field_dataptr_var = new ID(fmt("dataptr_after_%s", 
		getID()->Name()));
	env->AddID(end_of_field_dataptr_var, TEMP_VAR);

	genFieldEnd(out_cc, env, begin_ptr);

	end_of_field_dataptr = new DataPtr(env, end_of_field_dataptr_var, 0);
	return *end_of_field_dataptr;
	}

const char* Field::FieldSize(Output* out_cc, Env* env)
	{
	if ( field_size_expr )
		return field_size_expr;

	const DataPtr& begin = getFieldBegin(out_cc, env);
	const DataPtr& end = getFieldEnd(out_cc, env);
	if ( begin.getID() == end.getID() )
		field_size_expr = nfmt("%d", end.getOffset() - begin.getOffset());
	else
		field_size_expr = nfmt("(%s - %s)", end.PtrExpr(), begin.PtrExpr());
	return field_size_expr;
	}

const char* Field::FieldOffset(Output* out_cc, Env* env)
	{
	if ( field_offset_expr )
		return field_offset_expr;

	const DataPtr& begin = getFieldBegin(out_cc, env);
	if ( begin.getID() == &begin_of_data )
		field_offset_expr = nfmt("%d", begin.getOffset());
	else
		field_offset_expr = nfmt("(%s - %s)", 
			begin.PtrExpr(), env->RValue(&begin_of_data));
	return field_offset_expr;
	}

// The reasoning behind AttemptBoundaryCheck is: "If my next field
// can check its boundary, then I don't have to check mine, and it
// will save me a boundary-check."
bool Field::AttemptBoundaryCheck(Output* out_cc, Env* env)
	{
	if ( boundary_checked )
		return true;

	// If I do not even know my size till I parse the data, my
	// next field won't be able to check its boundary now.  

	const DataPtr& begin = getFieldBegin(out_cc, env);
	if ( StaticSize(env, begin.AbsOffset(&begin_of_data)) < 0 )
		return false;

	// Now we ask the next field to check its boundary. 
	if ( next && next->AttemptBoundaryCheck(out_cc, env) ) 
		{
		// If it works, we are all set
		SetBoundaryChecked();
		return true;
		}
	else
		// If it fails, then I can still try to do it by myself
		return genBoundaryCheck(out_cc, env);
	}

RecordField::RecordField(ID* arg_id, Type* arg_type)
	: Field(arg_id, RECORD), type(arg_type) 
	{
	type->SetValueVar(id, MEMBER_VAR);
	}

RecordField::~RecordField()
	{
	delete id;
	delete type;
	}

void RecordField::Prepare(Env* env)
	{
	type->Prepare(env, 0);
	env->SetEvalMethod(id, this);
	env->SetField(id, this);
	}

void RecordField::genPubDecls(Output* out_h, Env* env)
	{
	type->genPubDecls(out_h, env);
	}

void RecordField::genPrivDecls(Output* out_h, Env* env)
	{
	type->genPrivDecls(out_h, env);
	}

void RecordField::genInitCode(Output* out_cc, Env* env)
	{
	type->genInitCode(out_cc, env);
	}

void RecordField::genCleanUpCode(Output* out_cc, Env* env)
	{
	type->genCleanUpCode(out_cc, env);
	}

void RecordField::genParseCode(Output* out_cc, Env* env)
	{
	const DataPtr& data = getFieldBegin(out_cc, env);
	AttemptBoundaryCheck(out_cc, env);
	out_cc->println("// Parse \"%s\"", id->Name());

	type->AddSizeVar(out_cc, env);

	Env field_env(env);

	// Do we have a custom byteorder specified for this field?
	// (Or inherited from the upper level parsing?)
	LetDef* custom_byteorder = 0;
	foreach ( i, AttrList, attrlist )
		{
		Attr* a = *i;
		if ( a->Type() == ATTR_BYTEORDER )
			{
			AttrByteOrder* ab = (AttrByteOrder*) a;
			custom_byteorder = 
				new LetDef(&byteorder_id, 
					ab->IsLittleEndian());
			}
		}

	if ( custom_byteorder )
		{
		field_env.AddID(&byteorder_id, TEMP_VAR);
		field_env.SetEvaluated(&byteorder_id, false);
		field_env.SetEvalMethod(&byteorder_id, custom_byteorder);
		}

	type->genParseCode(out_cc, &field_env, data);

	SetBoundaryChecked();

	delete custom_byteorder;
	}

void RecordField::genEval(Output* out_cc, Env* env)
	{
	genParseCode(out_cc, env);
	}

void RecordField::genFieldEnd(Output* out_cc, Env* env, 
		const DataPtr& field_begin)
	{
	out_cc->println("const uint8* const %s = %s + (%s);", 
		env->LValue(end_of_field_dataptr_var),
		field_begin.PtrExpr(),
		type->DataSize(out_cc, env, field_begin));
	env->SetEvaluated(end_of_field_dataptr_var);
	}

void RecordField::SetBoundaryChecked()
	{
	Field::SetBoundaryChecked();
	type->SetBoundaryChecked();
	}

bool RecordField::genBoundaryCheck(Output* out_cc, Env* env)
	{
	if ( boundary_checked )
		return true;

	type->genBoundaryCheck(out_cc, env, getFieldBegin(out_cc, env));

	SetBoundaryChecked();
	return true;
	}

void LetDef::genEval(Output* out_cc, Env* env)
	{
	if ( ! env->Evaluated(id) )
		{
		out_cc->println("%s = %s;", 
			env->LValue(id), expr->EvalExpr(out_cc, env));
		env->SetEvaluated(id);
		}
	}

LetField::LetField(ID* arg_id, Expr* arg_expr)
	: Field(arg_id, LET), expr(arg_expr), let_def(id, expr)
	{
	type = new BasicType(BasicType::INT32);
	type->SetValueVar(id, MEMBER_VAR);
	}

LetField::~LetField()
	{
	delete type;
	delete expr;
	}

void LetField::Prepare(Env* env)
	{
	type->Prepare(env, 0);
	env->SetEvalMethod(id, &let_def);
	}

void LetField::genPubDecls(Output* out_h, Env* env)
	{
	out_h->println("%s const & %s const 	{ return %s; }", 
		expr->DataType(), env->RValue(id), env->LValue(id));
	}

void LetField::genPrivDecls(Output* out_h, Env* env)
	{
	out_h->println("%s %s;", expr->DataType(), env->LValue(id));
	}

void LetField::genInitCode(Output* out_cc, Env* env)
	{
	int v;
	if ( expr->ConstFold(env, v) )
		{
		DEBUG_MSG("Folding const for `%s'\n", id->Name());
		let_def.genEval(out_cc, env);
		}
	else
		type->genInitCode(out_cc, env);
	}

void LetField::genCleanUpCode(Output* out_cc, Env* env)
	{
	type->genCleanUpCode(out_cc, env);
	}

void LetField::genParseCode(Output* out_cc, Env* env)
	{
	let_def.genEval(out_cc, env);
	}

void LetField::genFieldEnd(Output* out_cc, Env* env, const DataPtr& field_begin)
	{
	// We do not need to compute the dynamic size of a "let"
	// field, because it always has static size 0.
	ASSERT(0);	
	out_cc->println("const uint8* const %s = %s;",
		env->LValue(end_of_field_dataptr_var),
		field_begin.PtrExpr());
	}

bool LetField::BoundaryChecked() const
	{
	if ( prev )
		return prev->BoundaryChecked();
	else
		return true;
	}

bool LetField::genBoundaryCheck(Output* out_cc, Env* env)
	{
	return false;
	}

PaddingField::PaddingField(ID* arg_id, PaddingType arg_type, Expr* arg_expr)
	: Field(arg_id, PADDING), type(arg_type), expr(arg_expr)
	{
	wordsize = -1;
	}

PaddingField::~PaddingField()
	{
	}

void PaddingField::Prepare(Env* env)
	{
	if ( type == PAD_TO_NEXT_WORD )
		{
		if ( ! expr->ConstFold(env, wordsize) )
			throw ExceptionPaddingError(this, 
				fmt("padding word size not a constant"));
		}
	}

void PaddingField::genParseCode(Output* out_cc, Env* env)
	{
	}

int PaddingField::StaticSize(Env* env, int offset) const
	{
	int length;
	int target_offset;
	int offset_in_word;

	switch ( type )
		{
		case PAD_BY_LENGTH:
			return expr->ConstFold(env, length) ? length : -1;

		case PAD_TO_OFFSET:
			// If the current offset cannot be statically
			// determined, we need to generate code to
			// check the offset
			if ( offset == -1 )
				return -1;

			if ( ! expr->ConstFold(env, target_offset) )
				return -1;

			// If both the current and target offsets
			// can be statically computed, we can get its
			// static size
			if ( offset > target_offset )
				throw ExceptionPaddingError(
					this,
					fmt("current offset = %d, "
					    "target offset = %d", 
					    offset, target_offset));
			return target_offset - offset;

		case PAD_TO_NEXT_WORD:
			if ( offset == -1 || wordsize == -1 )
				return -1;

			offset_in_word = offset % wordsize;
			return ( offset_in_word == 0 ) ? 
				0 : wordsize - offset_in_word;
		}

	return -1;
	}

void PaddingField::genFieldEnd(Output* out_cc, Env* env, const DataPtr& field_begin)
	{
	char* padding_var;
	switch ( type )
		{
		case PAD_BY_LENGTH:
			out_cc->println("const uint8* const %s = %s + (%s);",
				env->LValue(end_of_field_dataptr_var),
				field_begin.PtrExpr(),
				expr->EvalExpr(out_cc, env));
			break;

		case PAD_TO_OFFSET:
			out_cc->println("const uint8* %s = %s + (%s);",
				env->LValue(end_of_field_dataptr_var),
				env->RValue(&begin_of_data),
				expr->EvalExpr(out_cc, env));
			env->SetEvaluated(end_of_field_dataptr_var);
			out_cc->println("if ( %s < %s )",
				env->RValue(end_of_field_dataptr_var),
				field_begin.PtrExpr());
			out_cc->inc_indent();
			out_cc->println("{");
			out_cc->println("// throw ExceptionInvalidOffset(\"%s\", %s - %s, %s);",
				id->LocName(), 
				field_begin.PtrExpr(),
				env->RValue(&begin_of_data),
				expr->EvalExpr(out_cc, env));
			out_cc->println("%s = %s;", 
				env->RValue(end_of_field_dataptr_var),
				field_begin.PtrExpr());
			out_cc->println("}");
			out_cc->dec_indent();
			break;

		case PAD_TO_NEXT_WORD:
			padding_var = nfmt("%s__size", getID()->Name());
			out_cc->println("int %s = (%s - %s) %% %d;",
				padding_var, 
				field_begin.PtrExpr(),
				env->RValue(&begin_of_data),
				wordsize);
			out_cc->println("%s = (%s == 0) ? 0 : %d - %s;",
				padding_var,
				padding_var,
				wordsize,
				padding_var);
			out_cc->println("const uint8* const %s = %s + %s;",
				env->LValue(end_of_field_dataptr_var), 
				field_begin.PtrExpr(),
				padding_var);
			delete [] padding_var;
			break;
		}
	env->SetEvaluated(end_of_field_dataptr_var);
	}

bool PaddingField::genBoundaryCheck(Output* out_cc, Env* env)
	{
	if ( boundary_checked )
		return true;

	const DataPtr& begin = getFieldBegin(out_cc, env);

	char* size;
	int ss = StaticSize(env, begin.AbsOffset(&begin_of_data));
	ASSERT ( ss >= 0 );
 	size = nfmt("%d", ss);
	
	begin.genBoundaryCheck(out_cc, env, size, field_id_str);

	delete [] size;

	SetBoundaryChecked();
	return true;
	}

CaseType::CaseType(CaseIndex* arg_index, CaseList* arg_cases)
	: index(arg_index), cases(arg_cases) 
	{
	tot_ = CASE;
	index_var = 0;
	}

CaseType::~CaseType()
	{
	delete index;
	delete_list(CaseList, cases);
	}

const char* CaseType::DataType()
	{
	ASSERT(0);
	return 0;
	}

void CaseType::Prepare(Env* env, const char* enclosing_lvalue)
	{
	Type::Prepare(env, enclosing_lvalue);

	index->Prepare(env, value_var);
	index_var = index->IndexVar();
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		c->Prepare(env, value_var);
		c->SetIndexVar(index_var);
		}
	}

void CaseType::genPrivDecls(Output* out_h, Env* env)
	{
	index->genPrivDecls(out_h, env);
	out_h->println("union {");
	out_h->inc_indent();
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		c->genPrivDecls(out_h, env);
		}
	out_h->dec_indent();
	out_h->println("} %s;", env->LValue(value_var));
	}

void CaseType::genPubDecls(Output* out_h, Env* env)
	{
	index->genPubDecls(out_h, env);
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		c->genPubDecls(out_h, env);
		}
	}

void CaseType::genInitCode(Output* out_cc, Env* env)
	{
	index->genInitCode(out_cc, env);

#if 0
	ASSERT(env->Evaluated(index_var));
	out_cc->println("switch ( %s )", env->RValue(index_var));
	out_cc->inc_indent();
	out_cc->println("{");
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		c->genInitCode(out_cc, env);
		}
	out_cc->println("}");
	out_cc->dec_indent();
#endif
	}

void CaseType::genCleanUpCode(Output* out_cc, Env* env)
	{
	out_cc->println("switch ( %s )", env->RValue(index_var));
	out_cc->inc_indent();
	out_cc->println("{");
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		c->genCleanUpCode(out_cc, env);
		}
	out_cc->println("}");
	out_cc->dec_indent();

	index->genCleanUpCode(out_cc, env);
	}

void CaseType::do_genParseCode(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	if ( StaticSize(env) >= 0 )
		genBoundaryCheck(out_cc, env, data);

	AddSizeVar(out_cc, env);
	bool compute_size_var = (size_var && ! env->Evaluated(size_var));
	if ( compute_size_var )
		out_cc->println("int %s;", env->LValue(size_var));

	index->genParseCode(out_cc, env);
	out_cc->println("switch ( %s )", env->RValue(index_var));
	out_cc->inc_indent();
	out_cc->println("{");
	bool has_default_case = false;
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		c->genParseCode(out_cc, env, data, 
			compute_size_var ? size_var : 0);
		if ( c->IsDefaultCase() )
			has_default_case = true;
		}
	if ( ! has_default_case )
		{
		out_cc->println("default:");
		out_cc->inc_indent();
		out_cc->println("throw ExceptionInvalidCaseIndex(\"%s\", %s);", 
			decl_id->Name(), env->RValue(index_var));
		out_cc->println("break;");
		out_cc->dec_indent();
		}
	out_cc->println("}");
	out_cc->dec_indent();
	}

void CaseType::genDynamicSize(Output* out_cc, Env* env,
		const DataPtr& data)
	{
	genParseCode(out_cc, env, data);
	}

int CaseType::StaticSize(Env* env) const
	{
	int static_w = -1;
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		int w = c->StaticSize(env);
		if ( w < 0 || ( static_w >= 0 && w != static_w ) )
			return -1;
		static_w = w;
		}
	return static_w;
	}

void CaseType::SetBoundaryChecked()
	{
	Type::SetBoundaryChecked();
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		c->SetBoundaryChecked();
		}
	}

bool CaseType::ByteOrderSensitive() const
	{
	foreach (i, CaseList, cases)
		{
		Case* c = *i;
		if ( c->ByteOrderSensitive() )
			return true;
		}
	return false;
	}

CaseIndex::CaseIndex(Expr* arg_expr)
	: expr(arg_expr)
	{
	index_var = 0;
	}

CaseIndex::~CaseIndex()
	{
	delete index_var;
	delete expr;
	}

void CaseIndex::Prepare(Env* env, const ID* case_var)
	{
	index_var = new ID(fmt("%s_case_index", case_var->Name()));
	env->AddID(index_var, MEMBER_VAR);
	}

void CaseIndex::genPubDecls(Output* out_h, Env* env)
	{
	out_h->println("int %s const	{ return %s; }",
		env->RValue(index_var), env->LValue(index_var));
	}

void CaseIndex::genPrivDecls(Output* out_h, Env* env)
	{
	out_h->println("int %s;", env->LValue(index_var));
	}

void CaseIndex::genInitCode(Output* out_cc, Env* env)
	{
	out_cc->println("%s = -1;", env->LValue(index_var));
	}

void CaseIndex::genParseCode(Output* out_cc, Env* env)
	{
	out_cc->println("%s = %s;", 
		env->LValue(index_var), expr->EvalExpr(out_cc, env));
	env->SetEvaluated(index_var);
	}

void CaseIndex::genCleanUpCode(Output* out_cc, Env* env)
	{
	// do nothing
	}

Case::Case(Expr* arg_expr, ID* arg_id, Type* arg_type)
	: expr(arg_expr), id(arg_id), type(arg_type) 
	{
	type->SetValueVar(id, UNION_VAR);
	case_str = 0;
	lvalue = 0;
	}

Case::~Case()
	{
	delete expr;
	delete id;
	delete type;
	delete [] case_str;
	}

void Case::Prepare(Env* env, const ID* case_var)
	{
	type->Prepare(env, env->LValue(case_var));
	lvalue = type->LValue();
	if ( expr )
		{
		if ( ! expr->ConstFold(env, index_const) )
			throw ExceptionNonConstExpr(expr);
		case_str = nfmt("case %d:", index_const);
		}
	else
		{
		index_const = -1;
		case_str = nfmt("default:");
		}
	}

void Case::genPubDecls(Output* out_h, Env* env)
	{
	if ( ! type->DataType() )
		return;

	out_h->println("%s const & %s const",
		type->DataType(), env->RValue(id));

	out_h->inc_indent();
	out_h->println("{");

	if ( ! expr )
		out_h->println("return %s;", lvalue);
	else
		{
		out_h->println("if ( %s == %d )",
			env->RValue(index_var), index_const);
		out_h->inc_indent();
		out_h->println("return %s;", lvalue);
		out_h->dec_indent();

		out_h->println("else");
		out_h->inc_indent();
		out_h->println("throw ExceptionInvalidCase(\"%s\", %s, %d);", 
			id->LocName(), env->RValue(index_var), index_const);
		out_h->dec_indent();
		}

	out_h->println("}");
	out_h->dec_indent();
	}

void Case::genPrivDecls(Output* out_h, Env* env)
	{
	if ( ! type->DataType() )
		return;

	out_h->println("%s %s;", type->DataType(), env->LValue(id));
	}

void Case::genInitCode(Output* out_cc, Env* env)
	{
	out_cc->println("%s", case_str);
	out_cc->inc_indent();
	out_cc->println("// Initialize \"%s\"", id->Name());
	out_cc->println("{");
	type->genInitCode(out_cc, env);
	out_cc->println("}");
	out_cc->println("break;");
	out_cc->dec_indent();
	}

void Case::genCleanUpCode(Output* out_cc, Env* env)
	{
	out_cc->println("%s", case_str);
	out_cc->inc_indent();
	out_cc->println("// Clean up \"%s\"", id->Name());
	out_cc->println("{");
	type->genCleanUpCode(out_cc, env);
	out_cc->println("}");
	out_cc->println("break;");
	out_cc->dec_indent();
	}

void Case::genParseCode(Output* out_cc, Env* env, 
		const DataPtr& data, const ID* size_var)
	{
	out_cc->println("%s", case_str);
	out_cc->inc_indent();
	out_cc->println("// Parse \"%s\"", id->Name());
	out_cc->println("{");
	type->genParseCode(out_cc, env, data);
	if ( size_var )
		{
		out_cc->println("%s = %s;", 
			env->LValue(size_var),
			type->DataSize(out_cc, env, data));
		env->SetEvaluated(size_var);
		}
	out_cc->println("}");
	out_cc->println("break;");
	out_cc->dec_indent();
	}

SeqEnd::SeqEnd(NumList* arg_numlist)
	: numlist(arg_numlist)
	{
	ASSERT ( numlist != 0 );
	}

SeqEnd::~SeqEnd()
	{
	}

void SeqEnd::genParseCode(Output* out_cc, Env* env, 
		const ID* data_var, const char* seq_name)
	{
	int n = numlist->size();
	char* nn = nfmt("%d", n);
	
	// First check the data boundary
	DataPtr data(env, data_var, 0);
	data.genBoundaryCheck(out_cc, env, nn, seq_name);
	
	out_cc->println("if ( // Have we reached the end of the sequence?");
	for ( int i = 0; i < n; ++i )
		{
		out_cc->println("     %s[%d] == %s %s", 
			env->RValue(data_var), i,
			(*numlist)[i]->Str(), 
			( i < n - 1 ) ? "&&" : "");
		}
	out_cc->println("   )");
	out_cc->inc_indent();
	out_cc->println("{");
	out_cc->println("%s += %d;", env->LValue(data_var), n);
	out_cc->println("break;");
	out_cc->println("}");
	out_cc->dec_indent();

	delete [] nn;
	}

static const char* expr_fmt[] = 
{
#	define EXPR_DEF(type, num_op, fmt) fmt,
#	include "pac_expr.def"
#	undef EXPR_DEF
};

void Expr::init()
	{
	id = 0;
	num = 0;
	str = 0;
	orig = 0;
	num_operands = 0;
	operand[0] = 0;
	operand[1] = 0;
	operand[2] = 0;
	}

Expr::Expr(ID* arg_id)
	{
	init();
	type = EXPR_ID;
	id = arg_id;
	num_operands = 0;
	orig = nfmt("%s", id->Name());
	}

Expr::Expr(Number* arg_num)
	{
	init();
	type = EXPR_NUM;
	num = arg_num;
	num_operands = 0;
	orig = nfmt("((int) %s)", num->Str());
	}

Expr::Expr(expr_type arg_type, Expr* op1)
	{
	init();
	type = arg_type;
	num_operands = 1;
	operand[0] = op1;
	orig = nfmt(expr_fmt[type], op1->Orig());
	}

Expr::Expr(expr_type arg_type, Expr* op1, Expr* op2)
	{
	init();
	type = arg_type;
	num_operands = 2;
	operand[0] = op1;
	operand[1] = op2;
	operand[2] = 0;
	orig = nfmt(expr_fmt[type], op1->Orig(), op2->Orig());
	}

Expr::Expr(expr_type arg_type, Expr* op1, Expr* op2, Expr* op3)
	{
	init();
	type = arg_type;
	num_operands = 3;
	operand[0] = op1;
	operand[1] = op2;
	operand[2] = op3;
	orig = nfmt(expr_fmt[type], op1->Orig(), op2->Orig(), op3->Orig());
	}

Expr::~Expr()
	{
	delete [] str;
	delete id;
	delete operand[0];
	delete operand[1];
	delete operand[2];
	}

const char* Expr::ExprStr(Env* env) const
	{
	switch ( type ) 
		{
		case EXPR_NUM:	return num->Str();
		case EXPR_ID:	return env->RValue(id);
		default:	return str;
		}
	}

void Expr::genStr(Env* env)
	{
	if ( str )
		return;

	for ( int i = 0; i < 3; ++i )
		if ( operand[i] )
			operand[i]->genStr(env);

	switch ( num_operands )
		{
		case 1:
			str = nfmt(expr_fmt[type], 
				operand[0]->ExprStr(env));
			break; 
		case 2:
			str = nfmt(expr_fmt[type], 
				operand[0]->ExprStr(env),
				(type == EXPR_MEMBER) ? 
					operand[1]->getID()->Name() : 
					operand[1]->ExprStr(env));
			break; 
		case 3:
			str = nfmt(expr_fmt[type], 
				operand[0]->ExprStr(env),
				operand[1]->ExprStr(env),
				operand[2]->ExprStr(env));
			break; 
		}
	}

void Expr::genEval(Output* out_cc, Env* env)
	{
	switch ( type )
		{
		case EXPR_NUM:
			return;

		case EXPR_ID:
			env->Evaluate(out_cc, id);
			return;

		case EXPR_MEMBER:
			/*
			For member expressions such X.Y, evaluating
			X only is sufficient. (Actually trying to
			evaluate Y will lead to error because Y is
			not defined in the current environment.)
			*/
			operand[0]->genEval(out_cc, env);
			return;

		case EXPR_SIZEOF:
			{
			Field* field = env->getField(operand[0]->getID());
			if ( ! field )
				throw ExceptionIDNotField(operand[0]->getID());
			str = nfmt("%s", field->FieldSize(out_cc, env));
			return;
			}

		case EXPR_OFFSETOF:
			{
			Field* field = env->getField(operand[0]->getID());
			if ( ! field )
				throw ExceptionIDNotField(operand[0]->getID());
			str = nfmt("%s", field->FieldOffset(out_cc, env));
			return;
			}

		default:
			// Evaluate every operand by default
			for ( int i = 0; i < 3; ++i )
				if ( operand[i] )
					operand[i]->genEval(out_cc, env);
			break;
		}
	}

const char* Expr::EvalExpr(Output* out_cc, Env* env)
	{
	genEval(out_cc, env);
	genStr(env);
	return ExprStr(env);
	}

bool Expr::ConstFold(Env* env, int& n) const
	{
	switch ( type ) 
		{
		case EXPR_NUM:	
			n = num->Num();
			return true;
		case EXPR_ID:
			return env->getConstant(id, n);
		default:	
			// ### FIXME: folding consts
			return false;
		}
	}
