// $Id: pac.h,v 1.2 2005/09/07 17:33:56 vern Exp $

#ifndef pac_h
#define pac_h

#include <stdio.h>
#include <vector>
#include <string>
#include <map>

using namespace std;

#include "pac_dbg.h"
#include "pac_utils.h"
#include "pac_output.h"

extern const char* input_filename;
extern int line_number;

class Object
{
public:
	Object()
		{
		filename = input_filename;
		line_num = line_number;
		asprintf(&location, "%s:%d", input_filename, line_number);
		}

	~Object()
		{
		free(location);
		}

	const char* Location() const	{ return location; }

protected:
	const char* filename;
	int line_num;
	char* location;
};

class ID;
class Number;
class Decl;
class Type;
class Param;
class Field;
class Padding;
class CaseIndex;
class Case;
class SeqEnd;
class Expr;
class Attr;
class Enum;

typedef vector<Param*> 		ParamList;
typedef vector<Field*> 		FieldList;
typedef vector<Case*> 		CaseList;
typedef vector<Expr*> 		ExprList;
typedef vector<Enum*>		EnumList;
typedef vector<Number*>		NumList;
typedef vector<Attr*> 		AttrList;

#define foreach(i, ct, pc) \
	if ( pc ) \
		for ( ct::iterator i = (pc)->begin(); i != (pc)->end(); ++i )

#define delete_list(ct, pc) \
	foreach(delete_list_i, ct, pc) 		\
		delete *delete_list_i;		\
	delete pc;				\
	pc = 0;

class ID : public Object
{
public:
	ID(const char* arg_name)
		: name(arg_name) 
		{
		locname = nfmt("%s:%s", Location(), Name());
		}
	~ID()
		{
		delete locname;
		}
	const char* Name() const 	{ return name.c_str(); }
	const char* LocName() const 	{ return locname; }

protected:
	string name;
	char* locname;
	friend class ID_cmp;
};

class ID_cmp
{
public:
	bool operator()(const ID* const & id1, const ID* const & id2) const
		{
		return id1->name < id2->name;
		}
};

class Env;
class Evaluatable
{
public:
	virtual void genEval(Output* out, Env* env) = 0;
};

enum ID_type {
	CONST,
	GLOBAL_VAR,
	TEMP_VAR,
	MEMBER_VAR,
	UNION_VAR,
};

class IDRecord
{
public:
	IDRecord(const ID* id, ID_type id_type);
	~IDRecord();

	void SetEvalMethod(Evaluatable* arg_eval) { eval = arg_eval; }
	void Evaluate(Output* out, Env* env);
	void SetEvaluated(bool v)		{ evaluated = v; }
	bool Evaluated() const 			{ return evaluated; }

	void SetField(Field* f)			{ field = f; }
	Field* getField() const			{ return field; }

	void SetConstant(int c)			
		{ 
		constant_set = true;
		constant = c; 
		}
	bool getConstant(int& c) const
		{ 
		if ( constant_set ) 
			c = constant;
		return constant_set;
		}

	const char* RValue() const;
	const char* LValue() const		{ return lvalue; }

protected:
	const ID* id;
	ID_type id_type;

	char* rvalue;
	char* lvalue;

	Field* field;

	int constant;
	bool constant_set;

	bool evaluated;
	bool in_evaluation;	// to detect cyclic dependence
	Evaluatable* eval;
};

class Env
{
public:
	Env(Env* parent_env);
	~Env();

	void AddID(const ID* id, ID_type id_type);
	void AddConstID(const ID* id, const int c);

	const char* RValue(const ID* id) const;
	const char* LValue(const ID* id) const;

	void SetEvalMethod(const ID* id, Evaluatable* eval);
	void Evaluate(Output* out, const ID* id);
	bool Evaluated(const ID* id) const;
	void SetEvaluated(const ID* id, bool v = true);

	void SetField(const ID* id, Field* field);
	Field* getField(const ID* id) const;

	void SetConstant(const ID* id, int constant);
	bool getConstant(const ID* id, int& c) const;

protected:
	IDRecord* lookup(const ID* id, 
		bool recursive = true, bool raise_exception = true) const;

	Env* parent;
	typedef map<const ID*, IDRecord*, ID_cmp> id_map_t;
	id_map_t id_map;
};

class Number
{
public:
	Number(int arg_n)
		: s(fmt("%d", arg_n)), n(arg_n) {}
	Number(const char* arg_s, int arg_n)
		: s(arg_s), n(arg_n) {}
	const char* Str() const 	{ return s.c_str(); }
	const int Num() const 		{ return n; }

protected:
	const string s;
	const int n;
};

// A data pointer is represented by an data pointer variable
// plus a constant offset.

class DataPtr
{
public:
	DataPtr(Env* env, const ID* arg_id, const int arg_off);
	~DataPtr() 			{ delete [] ptr_expr; }

	const ID* getID() const 	{ return id; }
	int getOffset() const 		{ return off; }

	const char* PtrExpr() const	{ return ptr_expr; }

	int AbsOffset(const ID* base_ptr) const;
	char* AbsOffsetExpr(Env* env, const ID* base_ptr) const;

	void genBoundaryCheck(Output* out, Env* env, 
		const char* data_size, const char* data_name) const;

protected:
	const ID* const id;
	const int off;
	char* ptr_expr;
};

class LetDefinition 
{
public:
	LetDefinition(const ID* arg_id, Expr* arg_expr);
	void genEval(Output* out, Env* env);

protected:
};

class Decl : public Object
{
public:
	Decl(ID* id);
	virtual ~Decl();

	virtual void genCode(Output* out_h, Output* out_cc) = 0;

	void AddAttr(AttrList* attrlist);

protected:
	virtual void ProcessAttr(Attr* a) {}

	ID* id;
	AttrList* attrlist;
};

extern const ID* current_decl_id;

class Enum
{
public:
	Enum(const ID* arg_id, Expr* arg_expr = 0);
	void genHeader(Output* out_h, int& c);

protected:
	const ID* id;
	Expr* expr;
};

class EnumDecl : public Decl
{
public:
	EnumDecl(ID* arg_id, EnumList* arg_enumlist)
		: Decl(arg_id), enumlist(arg_enumlist) {}
	~EnumDecl()
		{ delete_list(EnumList, enumlist); }
	void genCode(Output* out_h, Output* out_cc);

protected:
	EnumList* enumlist;
};

class LetDecl : public Decl
{
public:
	LetDecl(ID* id, Expr* expr);
	~LetDecl();
	void genCode(Output* out_h, Output* out_cc);

protected:
	Type* type;
	Expr* expr;
};

class TypeDecl : public Decl
{
public:
	TypeDecl(ID* arg_id, ParamList* arg_params, Type* arg_type);
	~TypeDecl();
	void genCode(Output* out_h, Output* out_cc);
	void genParamPubDecls(Output* out);
	void genParamPrivDecls(Output* out);

protected:
	void ProcessAttr(Attr* a);
	const char* ClassName() { return id->Name(); }

	void Prepare();

	char* ConstructorFuncPrototype(Env* env);
	char* DestructorFuncPrototype(Env* env);
	char* ParseFuncPrototype(Env* env);

	void genConstructorFunc(Output* out_h, Output* out_cc);
	void genDestructorFunc(Output* out_h, Output* out_cc);
	void genParseFunc(Output* out_h, Output* out_cc);

	Env* env;

	ParamList* params;
	Type* type;
	Expr* byteorder_expr;
};

class Param : public Object
{
public:
	Param(ID* arg_id, ID* arg_type_id = 0); 
	~Param();

	const char* Decl() const { return decl_str; }
	const ID* getID() const { return id; }
	const char* getTypeName() const { return type_name; }
	const char* Name() const { return id->Name(); }

protected:
	ID* id;
	ID* type_id;
	char* decl_str;
	char* type_name;
};

class Type : public Object
{
public:
	enum type_of_type {
		UNDEF = -1,
		EMPTY,
		BASIC,
		PARAMETERIZED,
		RECORD,
		CASE,
		ARRAY,
		SEQ,
	};

	Type();
	virtual ~Type();

	const ID* DeclID() const		{ return decl_id; }
	void SetDeclID(const ID* arg_id) 	{ decl_id = arg_id; }

	void SetValueVar(const ID* arg_id, ID_type arg_id_type)
	 	{ value_var = arg_id; value_var_type = arg_id_type; }

	const type_of_type tot() { return tot_; }
	virtual const char* DataType() = 0;

	virtual void Prepare(Env* env, const char* enclosing_lvalue);

	virtual void genPubDecls(Output* out, Env* env);
	virtual void genPrivDecls(Output* out, Env* env);

	virtual void genInitCode(Output* out, Env* env);
	virtual void genCleanUpCode(Output* out, Env* env);
	bool needsCleanUp() const;

	void genParseCode(Output* out, Env* env, const DataPtr& data);
	virtual void do_genParseCode(Output* out, Env* env,
		const DataPtr& data) = 0;

	virtual int StaticSize(Env* env) const = 0;
	const char* DataSize(Output* out, Env* env, const DataPtr& data);

	virtual void SetBoundaryChecked() 	{ boundary_checked = true; }
	void genBoundaryCheck(Output* out, Env* env, const DataPtr& data);

	bool AddSizeVar(Output* out, Env* env);
	virtual bool ByteOrderSensitive() const = 0;

	const char* LValue() const		{ return lvalue; }

	void AddCheck(Expr* expr)		{ /* TODO */ }

protected:
	type_of_type tot_;
	const ID* decl_id;
	const ID* value_var;
	char* data_id_str;
	ID_type value_var_type;
	ID* size_var;
	char* size_expr;
	bool boundary_checked;
	char* lvalue;

	virtual void genDynamicSize(Output* out, Env* env,
		const DataPtr& data) = 0;
};

class BasicType : public Type
{
public:
	enum basic_type {
#		define TYPE_DEF(name, pactype, ctype, size)	name,
#		include "pac_type.def"
#		undef TYPE_DEF
	};

	static int lookup_basictype(const char* name);

	BasicType(basic_type arg_type)
		: type(arg_type) 
		{
		if ( type == BasicType::EMPTY )
			tot_ = Type::EMPTY;
		else
			tot_ = Type::BASIC;
		}

	const char* DataType();
	void do_genParseCode(Output* out, Env* env, const DataPtr& data);

	int StaticSize(Env* env) const;

	bool ByteOrderSensitive() const { return StaticSize(0) >= 2; }

protected:
	basic_type type;

	void genDynamicSize(Output* out, Env* env, const DataPtr& data);
};

class ParameterizedType : public Type
{
public:
	ParameterizedType(ID* arg_type_id, ExprList* arg_args);
	~ParameterizedType();

	const char* DataType();

	void genCleanUpCode(Output* out, Env* env);
	void do_genParseCode(Output* out, Env* env, const DataPtr& data);

	int StaticSize(Env* env) const;
	bool ByteOrderSensitive() const;

protected:
	ID* type_id;
	ExprList* args;
	char* data_type;

	void genDynamicSize(Output* out, Env* env, const DataPtr& data);
};

class ArrayType : public Type
{
public:
	ArrayType(Type* arg_elemtype, Expr* arg_length);
	ArrayType(Type* arg_elemtype, SeqEnd* arg_end);
	~ArrayType();

	const char* DataType();

	void Prepare(Env* env, const char* enclosing_lvalue);

	void genPubDecls(Output* out, Env* env);
	void genPrivDecls(Output* out, Env* env);

	void genInitCode(Output* out, Env* env);
	void genCleanUpCode(Output* out, Env* env);

	void do_genParseCode(Output* out, Env* env, const DataPtr& data);

	int StaticSize(Env* env) const;

	void SetBoundaryChecked();

	bool ByteOrderSensitive() const 
		{ 
		return elemtype->ByteOrderSensitive(); 
		}

protected:
	void init();
	void genDynamicSize(Output* out, Env* env, const DataPtr& data);
	void genArrayLength(Output* out_cc, Env* env);

	Type* elemtype;
	Expr* length;
	SeqEnd* end;

	char* data_type;
	ID* arraylength_var;
	ID* elem_var;
	ID* elem_data_var;
};

class SeqEnd : public Object
{
public:
	SeqEnd(NumList* arg_numlist);
	~SeqEnd();

	void genParseCode(Output* out_cc, Env* env, 
		const ID* data_var, const char* seq_name);

	int StaticSize(Env* env) const	{ return numlist->size(); }

protected:
	NumList* numlist;
};

class RecordType : public Type
{
public:
	RecordType(FieldList* arg_fields);
	~RecordType();

	const char* DataType();

	void Prepare(Env* env, const char* enclosing_lvalue);

	void genPubDecls(Output* out, Env* env);
	void genPrivDecls(Output* out, Env* env);

	void genInitCode(Output* out, Env* env);
	void genCleanUpCode(Output* out, Env* env);
	void do_genParseCode(Output* out, Env* env, const DataPtr& data);

	int StaticSize(Env* env) const;

	void SetBoundaryChecked();

	bool ByteOrderSensitive() const;

protected:
	FieldList* fields;

	void genDynamicSize(Output* out, Env* env, const DataPtr& data);
};

class Field : public Object
{
public:
	enum FieldType {
		RECORD, LET, PADDING,
	};

	Field(ID* id, FieldType tof);
	virtual ~Field();

	void SetRecordType(RecordType* ty) 	{ record_type = ty; }
	void AddAttr(AttrList* attrlist);

	virtual void Prepare(Env* env) = 0;

	virtual void genPubDecls(Output* out, Env* env) = 0;
	virtual void genPrivDecls(Output* out, Env* env) = 0;

	virtual void genInitCode(Output* out, Env* env) = 0;
	virtual void genCleanUpCode(Output* out, Env* env) = 0;
	virtual void genParseCode(Output* out, Env* env) = 0;

	const FieldType tof() const 	{ return tof_; }
	const ID* getID() const		{ return id; }
	void SetPrev(Field* f) 		{ prev = f; }
	void SetNext(Field* f) 		{ next = f; }

	virtual int StaticSize(Env* env, int offset) const = 0;
	const char* FieldSize(Output* out, Env* env);
	const char* FieldOffset(Output* out, Env* env);

	virtual bool BoundaryChecked() const	{ return boundary_checked; }
	virtual void SetBoundaryChecked()	{ boundary_checked = true; }

	virtual bool ByteOrderSensitive() const = 0;

protected:
	ID* id;
	FieldType tof_;
	RecordType* record_type;
	const ID* decl_id;
	char* field_id_str;
	AttrList* attrlist;
	Field* prev;
	Field* next;
	bool boundary_checked;

	DataPtr* begin_of_field_dataptr;
	DataPtr* end_of_field_dataptr;
	char* field_size_expr;
	char* field_offset_expr;
	ID* end_of_field_dataptr_var;

	const DataPtr& Field::getFieldBegin(Output* out_cc, Env* env);
	const DataPtr& Field::getFieldEnd(Output* out_cc, Env* env);
	virtual void genFieldEnd(Output* out, Env* env, const DataPtr& begin) = 0; 

	bool AttemptBoundaryCheck(Output* out_cc, Env* env);
	virtual bool genBoundaryCheck(Output* out_cc, Env* env) = 0;

	friend class RecordType;
};

class RecordField : public Field, public Evaluatable
{
public:
	RecordField(ID* arg_id, Type* arg_type);
	~RecordField();

	void Prepare(Env* env);

	void genPubDecls(Output* out, Env* env);
	void genPrivDecls(Output* out, Env* env);

	void genInitCode(Output* out, Env* env);
	void genCleanUpCode(Output* out, Env* env);
	void genParseCode(Output* out, Env* env);

	void genEval(Output* out, Env* env);

	const ID* getID() const 	{ return id; }

	int StaticSize(Env* env, int) const 	{ return type->StaticSize(env); }

	void SetBoundaryChecked();

	bool ByteOrderSensitive() const	{ return type->ByteOrderSensitive(); }

protected:
	Type* type;

	void genFieldEnd(Output* out, Env* env, const DataPtr& begin); 
	bool genBoundaryCheck(Output* out_cc, Env* env);
};

class LetDef : public Evaluatable
{
public:
	LetDef(const ID* arg_id, Expr* arg_expr)
		: id(arg_id), expr(arg_expr) {}
	void genEval(Output* out_cc, Env* env);

protected:
	const ID* id;
	Expr* expr;
};

class LetField : public Field
{
public:
	LetField(ID* arg_id, Expr* arg_expr);
	~LetField();

	void Prepare(Env* env);

	void genPubDecls(Output* out, Env* env);
	void genPrivDecls(Output* out, Env* env);

	void genInitCode(Output* out, Env* env);
	void genCleanUpCode(Output* out, Env* env);
	void genParseCode(Output* out, Env* env);

	const ID* getID() const { return id; }
	int StaticSize(Env*, int) const 	{ return 0; }

	// As we do not check boundary for "let" fields, the return
	// value of BoundaryChecked() is computed differently.
	bool BoundaryChecked() const;

	bool ByteOrderSensitive() const	{ return false; }

protected:
	void genFieldEnd(Output* out, Env* env, const DataPtr& begin); 
	bool genBoundaryCheck(Output* out_cc, Env* env);

	Type* type;
	Expr* expr;

	LetDef let_def;
};

enum PaddingType { PAD_BY_LENGTH, PAD_TO_OFFSET, PAD_TO_NEXT_WORD };

class PaddingField : public Field
{
public:
	PaddingField(ID* arg_id, PaddingType arg_type, Expr* arg_expr);
	~PaddingField();

	void Prepare(Env* env);

	void genPubDecls(Output* out, Env* env) 		{ /* nothing */ }
	void genPrivDecls(Output* out, Env* env) 		{ /* nothing */ }

	void genInitCode(Output* out, Env* env) 		{ /* nothing */ }
	void genCleanUpCode(Output* out, Env* env)	{ /* nothing */ }
	void genParseCode(Output* out, Env* env);

	int StaticSize(Env* env, int offset) const;

	bool ByteOrderSensitive() const	{ return false; }

protected:
	PaddingType type;
	Expr* expr;
	int wordsize;

	void genFieldEnd(Output* out, Env* env, const DataPtr& begin); 
	bool genBoundaryCheck(Output* out_cc, Env* env);
};

class CaseType : public Type
{
public:
	CaseType(CaseIndex* arg_index, CaseList* arg_cases);
	~CaseType();

	const char* DataType();

	void Prepare(Env* env, const char* enclosing_lvalue);

	void genPubDecls(Output* out, Env* env);
	void genPrivDecls(Output* out, Env* env);

	void genInitCode(Output* out, Env* env);
	void genCleanUpCode(Output* out, Env* env);
	void do_genParseCode(Output* out, Env* env, const DataPtr& data);

	int StaticSize(Env* env) const;

	void SetBoundaryChecked();

	bool ByteOrderSensitive() const;

protected:
	CaseIndex* index;
	CaseList* cases;
	const ID* index_var;

	void genDynamicSize(Output* out, Env* env, const DataPtr& data);
};

class CaseIndex : public Object
{
public:
	CaseIndex(Expr* arg_expr);
	~CaseIndex();

	void Prepare(Env* env, const ID* case_var);

	void genPubDecls(Output* out, Env* env);
	void genPrivDecls(Output* out, Env* env);

	void genInitCode(Output* out, Env* env);
	void genCleanUpCode(Output* out, Env* env);

	void genParseCode(Output* out_cc, Env* env);

	const ID* IndexVar() const 	{ return index_var; }

protected:
	Expr* expr;
	ID* index_var;
};

class Case : public Object
{
public:
	Case(Expr* arg_expr, ID* arg_id, Type* arg_type);
	~Case();
	
	void Prepare(Env* env, const ID *case_var);

	void genPubDecls(Output* out, Env* env);
	void genPrivDecls(Output* out, Env* env);

	void genInitCode(Output* out, Env* env);
	void genCleanUpCode(Output* out, Env* env);
	void genParseCode(Output* out, Env* env, 
		const DataPtr& data, const ID* size_var);

	int StaticSize(Env* env) const 	{ return type->StaticSize(env); }
	void SetIndexVar(const ID* var) 	{ index_var = var; }

	bool IsDefaultCase() const	{ return ! expr; }
	void SetBoundaryChecked()	{ type->SetBoundaryChecked(); }

	bool ByteOrderSensitive() const	{ return type->ByteOrderSensitive(); }

protected:
	Expr* expr;
	ID* id;
	Type* type;
	const ID* index_var;
	int index_const;
	char* case_str;
	const char* lvalue;
};


class Expr : public Object
{
public:
	enum expr_type {
#		define EXPR_DEF(type, x, y) type,
#		include "pac_expr.def"
#		undef EXPR_DEF
	};

	void init();
	Expr(ID* id);
	Expr(Number* num);
	Expr(expr_type type, Expr* op1);
	Expr(expr_type type, Expr* op1, Expr* op2);
	Expr(expr_type type, Expr* op1, Expr* op2, Expr* op3);
	~Expr();

	const char* Orig() const	{ return orig; }
	const ID* getID() const 	{ return id; }
	const char* DataType() const 	{ return "int"; }
	const char* EvalExpr(Output* out, Env* env);
	bool ConstFold(Env* env, int& n) const;

protected:
	expr_type type;
	ID* id;
	Number* num;
	int num_operands;
	Expr* operand[3];
	char* str;
	char* orig;

	const char* ExprStr(Env* env) const;
	void genEval(Output* out, Env* env);
	void genStr(Env* env);
};

enum AttrType {
	ATTR_BYTEORDER,
	ATTR_CHECK,
};

class Attr
{
public:
	Attr(AttrType arg_type)
		: type(arg_type) {}

	AttrType Type() const 	{ return type; }

protected:
	AttrType type;
};

class AttrByteOrder : public Attr
{
public:
	AttrByteOrder(Expr* arg_is_little_endian)
		: Attr(ATTR_BYTEORDER), 
		  is_little_endian(arg_is_little_endian) {}

	Expr* IsLittleEndian() const { return is_little_endian; }

protected:
	Expr* is_little_endian;
};

class AttrCheck : public Attr
{
public: 
	AttrCheck(Expr* arg_expr)
		: Attr(ATTR_CHECK), expr(arg_expr) {}

	Expr* getExpr() { return expr; }

protected:
	Expr* expr;
};

int pac_init();

#endif /* pac_h */
